# 七、错误和恐慌

在这一章中，我们将深入探讨 Go 的错误检测和恢复特性，以及它们与 Java 方法的不同之处。当你完成了这一章，你应该能够清楚地识别 Go 和 Java 错误方法之间的相似和不同之处。

代码，尤其是函数中的代码，可以通过几种方式退出:

1.  成功-功能如预期完成。

2.  失败——由于某种可预见的情况，功能没有按预期完成。

3.  严重故障(又名死机)——由于一些意外或异常情况或错误代码，功能没有完成。

在像 Java 这样的每个函数只有一个返回值的语言中，情况 1 和情况 2 通常是结合在一起的，由返回值本身来决定。考虑一下`String.indexOf`函数，它返回目标的索引或者值< 0 来表示没有找到目标。对于返回对象的函数，通常会返回`null`来表示失败(如果`null`是合法的值，这就有问题了)。这通常是许多 NullPointerExceptions 的原因。

## Go 错误

Go 函数可以返回零个或多个结果。许多 Go 函数(至少)会返回一个错误值。这是一个常见的例子:

```go
func DoSomething() (err error) { ... }

```

这表示 *DoSomething* 函数可以返回一个`error`(一个内置的 Go 接口类型)，在这种情况下(按照惯例和习惯用法)命名为`err`。`err`值可以是`nil`或某个`error`实例。一个更完整的例子:

```go
func DoSomething() (err error) {
     :
     err = DoSomePart()
     if err != nil {
          return
     }
     :
     return
}

```

Go 有一种常用的、不太冗长的方法来编码这种模式，它结合了赋值和 if 测试:

```go
if err = DoSomePart(); err != nil {
     return
}

```

每个可能失败的函数都遵循这种模式。虽然比利用异常来报告故障的典型 Java 代码冗长得多，但这遵循了 Go 使用的更透明/明显的风格。

请注意，返回没有显式值。这是可行的，因为返回值被命名为`err`并且`err`被赋值。另一种选择(作者不太喜欢)是

```go
func DoSomething() error {
     :
     xxx := DoSomePart()  // unconventional name
     if xxx != nil {
          return xxx           // explicitly returned
     }
     :
     return xxx
}

```

在大多数情况下，Go 更喜欢从函数中返回一个错误值。这种模式在 Java 中通常被认为是不好的做法，因为它迫使调用者测试返回的错误。在 Go 中，这种模式被认为是最佳实践；程序员必须记住测试返回的错误。这是 Go 和 Java 编程风格之间的一个主要区别，许多刚开始用 Go 编程的 Java 程序员在习惯上有一些困难。

对于一些简单的函数，只需要一个成功/失败指示器就足够了，返回的错误值由一个布尔值代替。这通常是 Go 内置操作的情况，比如映射查找和类型断言。

## 陷入恐慌

在 Java 中，更严重的故障是通过抛出某个*异常*来指示的。对于什么时候应该抛出异常，什么时候应该返回错误，经常会有混淆(例如，当读取超过文件末尾时)，Java(和许多社区)库代码做出这种不一致的选择。

Go 通过使用总是返回一个错误值作为最后(或唯一)返回值的多值函数，使这种行为更加一致。对照`nil`测试错误值，以确定是否出现错误。一般来说，任何其他返回值只有在没有错误的情况下才有意义。只有当函数灾难性地失败时(内存不足、被零除、索引越界、参数无效等)。)是一个恐慌提出来的。

Java 支持异常的概念(从技术上来说 *Throwables* ，它是异常的超类)。异常是一个对象，当一个意外的/不寻常的情况出现时，它可以被抛出。一个例子是当零被用作除数时 JVM 抛出的 DivideByZeroException。另一个更严重的例子是当 JVM 不能满足一个`new`操作时抛出的 *OutOfMemoryError* 。Java 进程在`try`语句的`catch`块中抛出可抛出对象。可抛出的实例在 Java 代码中普遍被抛出和捕获。

Go 有一个类似但不太常用的概念，叫做*恐慌*。死机很像一个 throwable，可以由代码(您的或某个库)通过使用 Go 内置`panic(<value>)`函数来引发。该值可以是任何类型(但通常是一个`string`或(优选的)一个`error`实例)；不应使用`nil`值。

Go 代码引起恐慌的情况应该很少。在大多数情况下，代码应该返回一个错误。只有在无法预料的情况下才应该使用 panic，因为通过错误来报告它们会很麻烦，例如 Java 的 OutOfMemoryError 的 Go 等价物。

Go 不像 Java 那样有异常类型。相反，它有恐慌参数(更像是 Java *错误*混合了一些*运行时异常*)。Go 对 RuntimeException 和非 RuntimeException throwables 之间的 Java 区别没有概念。所有这些都映射到单个紧急值。永远不要声明函数可能引发的恐慌参数。

Java 有`try/finally`和`try/catch/finally`语句集。Go 不会。它使用*延迟的*函数来达到`finally`的效果。Go 使用一种不同但相似的机制来捕捉恐慌。

与 Java 非常相似，如果没有被捕获，恐慌通常会导致程序在打印回溯后退出。为了捕捉 Go 中的异常，可以使用内置的`recover()`函数，该函数返回与最近的异常一起发送的值(在特定的 goroutine 中)。为此，必须在已经延迟的函数中调用`recover()`。

就像 Java `catch`子句可以检查抛出的异常一样，deferred 函数可以检查值，进行一些更正，然后再次返回或引发异常。像在 Java 中一样，延迟函数可以在当前调用栈的任何地方。这里有一个简单的例子:

```go
func DoIt() (err error) {
     defer func() {
          p := recover()
          if p != nil  {  // a panic occurred
               // process the panic by (say) testing p value
               err = nil  // make containing function not return an error
          }
     }()
     :
      // any code that can panic
     if err != nil {
          panic(errors.New(fmt.Sprintf("panic: %v", err)))
            // or equivalently
           panic(fmt.Errorf("panic: %v", err))
     }
     :
     return
}

```

一般来说，Go 库和 Go 运行时避免引起恐慌。您的代码还应该。一种常见的情况是利用恐慌。如果一个函数得到一个非法的参数值，它通常会被报告为死机而不是错误返回。这种情况被认为是编程错误，而不是代码应该从中恢复的情况。注意，不是所有的地鼠都遵循这种方法，因此不会验证参数并产生恐慌；其他一些问题通常会在以后出现。该代码依赖于被提供有效的输入。

注意通常应该避免在死机恢复延迟函数中引起新的死机。这就像在 Java 中避免在`catch`或`finally`子句中抛出异常一样。

捕捉恐慌的一个关键区域是 goroutines。goroutine 中未处理的死机会导致 Go 程序崩溃。所以，最好不要让它们发生。这需要系统的纪律。为了实现这一点，作者建议所有的 goroutines 都由一个助手函数创建，类似于清单 [7-1](#PC6) 。

```go
package main

import (
     "errors"
     "fmt"
     "time"
)
var NoError = errors.New("no error")  // special error

func GoroutineLauncher(gr func(), c *(chan error)) {
     go func(){
          defer func(){
               if p := recover(); p != nil {
                    if c != nil {
                         // ensure we send an error
                         if err, ok := p.(error); ok {
                              *c <- err
                              return
                         }
                         *c <- errors.New(fmt.Sprintf("%v", p))
                    }
                    return
               }
               if c != nil {
                    *c <- NoError  // could also send nil and test for it
               }
          }()
          gr()
     }()
}

var N = 5

func main() {
     var errchan = make(chan error, N)  // N >= 1 based on max active goroutines
     // :
     GoroutineLauncher (func(){
          time.Sleep(2 * time.Second)  // simulate complex work
          panic("panic happened!")
     }, &errchan)
     // :
     time.Sleep(5 * time.Second)        // simulate other work
     // :
     err := <- errchan  // wait for result
     if err != NoError {
          fmt.Printf("got %q" , err.Error())
     }
}

Listing 7-1Capture Panics in a Goroutine Launcher Function

```

请注意，如果客户端不需要错误报告，可以省略错误通道。

这在运行时产生

```go
got "panic happened!"

```

## 图示错误和混乱

内置`error`类型简单。很多第三方包都扩展了它，比如 *JuJu Errors* 。 <sup>[1](#Fn1)</sup> 清单 [7-2](#PC8) 、 [7-3](#PC11) 和 [7-4](#PC15) 是如何扩展它的一些可能的例子。例如，收集多次出现的错误(比如在处理切片的元素时)。

```go
type MultError []error

func (me MultError) Error() (res string) {
     res = "MultError"
     sep := " "
     for _, e := range me {
          res = fmt.Sprintf("%s%s%s", res, sep, e.Error())
          sep = "; "
     }
     return
}
func (me MultError) String() string {
     return me.Error()
}

Listing 7-2Multiple Cause Errors

```

当被使用时

```go
me  := MultError(make([]error,0, 10))
for _, v := range []string{"one", "two", "three"} {
     me = append(me, errors.New(v))
}
fmt.Printf("MultipleError error: %s\n", me.Error())
fmt.Printf("MultipleError value: %v\n\n", me)

```

生产

```go
MultipleError error: MultError one; two; three
MultipleError value: MultError one; two; three

```

或者一个错误是由另一个错误引起的(很像 Java 中所有的 *Throwables* 都有原因)。

```go
type ErrorWithCause struct {
     Err   error
     Cause error
}

func NewError(err error) *ErrorWithCause {
     return NewErrorWithCause(err, nil)
}
func NewErrorWithCause(err error, cause error) *ErrorWithCause {
     if err == nil {
          err = errors.New("no error supplied")
     }
     return &ErrorWithCause{err, cause}
}
func (wc ErrorWithCause) Error() string {
     xerr := wc.Err
     xcause := wc.Cause
     if xcause == nil {
          xcause = errors.New("no root cause supplied")
     }
     return fmt.Sprintf("ErrorWithCause{%v %v}", xerr, xcause)
}
func (wc ErrorWithCause) String() string {
     return wc.Error()
}

Listing 7-3Error with a Cause

```

当被使用时

```go
fmt.Printf("ErrorWithCause error: %s\n", ewc.Error())
fmt.Printf("ErrorWithCause value: %v\n\n", ewc)

```

生产

```go
ErrorWithCause error: ErrorWithCause{error cause}
ErrorWithCause value: ErrorWithCause{error cause}

```

注意，如下所示的方法使得任何数据类型都可以充当`error`:

```go
func (x <sometype>) Error() string

```

这是因为`error`类型被有效地定义为

```go
type error interface {
     Error() string
}

```

Go `errors` <sup>[2](#Fn2)</sup> 软件包有几个有用的实用函数:

`errors.Is(<error>, <type>)`–解开错误，直到它与提供的类型匹配，如果找到则返回`true`。

`errors.As(<error>, <*type>)`–展开错误，直到它与提供的变量类型匹配，将错误转换为该类型，设置变量，如果找到，则返回`true`。

`errors.Unwrap(<error>)`–返回任何包装的错误(类似于 Java 异常的任何原因)；实际的错误类型必须有一个`Unwrap(<error>)`方法。

在 Go 中模拟 Java 异常行为是可能的。例如，为了引入类似于 Try/Catch/Finally 的行为，可以实现如下的小库。这里，Go 函数取代了 Java Try/Catch、Try/Finally 和 Try/Catch/Finally 语句。

每个 function 子句都是作为(典型的)函数文字提供的。没有像 Java 中那样对每个异常类型都进行捕获，因为 Go 对所有问题都只有一个单一的异常。整体函数返回 try 子句的错误。因为 try 和 catch 子句可能有错误，所以有时会返回错误对类型`TryCatchError`。

注意直接在延迟函数中而不是在`triageRecover(...)`函数中发出`recover()`函数是很重要的。

```go
type TryFunc func() error
type CatchFunc func(error) (rerr error, cerr error)
type FinallyFunc func()

type TryCatchError struct {
     tryError   error
     catchError error
}

func (tce *TryCatchError) Error() string {
     return tce.String()
}
func (tce *TryCatchError) String() string {
     return fmt.Sprintf("TryCatchError[%v %v]", tce.tryError, tce.catchError)
}
func (tce *TryCatchError) Cause() error {
     return tce.tryError
}
func (tce *TryCatchError) Catch() error {
     return tce.catchError
}

func TryFinally(t TryFunc, f FinallyFunc) (err error) {
     defer func() {
          f()
     }()
     err = t()
     if err != nil {
          err = &TryCatchError{err, nil}
     }
     return
}

func triageRecover(p interface{}, c CatchFunc) (err error) {
     if p != nil {
          var terr, cerr error
          if v, ok := p.(error); ok {
               terr = v
          }
          if xrerr, xcerr := c(terr); xrerr != nil {
               cerr = xcerr
               err = xrerr
          }
          if terr != nil || cerr != nil {
               err = &TryCatchError{terr, cerr}
          }
     }
     return err
}

func TryCatch(t TryFunc, c CatchFunc) (err error) {
     defer func() {
          if xerr := triageRecover(recover(), c); xerr != nil {
               err = xerr
          }
     }()
     err = t()
     return
}
func TryCatchFinally(t TryFunc, c CatchFunc, f FinallyFunc) (err error) {
     defer func() {
          f()
     }()
     defer func() {
          if xerr := triageRecover(recover(), c); xerr != nil {
               err = xerr
          }
     }()
     err = t()
     return
}

Listing 7-4Try/Catch Emulation Example (Part 1)

```

这可以如清单 [7-5](#PC16) 所示使用。

```go
err := TryCatchFinally(func() error {
     fmt.Printf("in try\n")
     panic(errors.New("forced panic"))
}, func(e error) (re, ce error) {
     fmt.Printf("in catch %v: %v %v\n", e, re, ce)
     return
}, func() {
     fmt.Printf("in finally\n")
})
fmt.Printf("TCF returned: %v\n", err)

err = TryFinally(func() error {
     fmt.Printf("in try\n")
     return errors.New("try error")
}, func() {
     fmt.Printf("in finally\n")
})
fmt.Printf("TCF returned: %v\n", err)

err = TryCatch(func() error {
     fmt.Printf("in try\n")
     panic(errors.New("forced panic"))
}, func(e error) (re, ce error) {
     fmt.Printf("in catch %v: %v %v\n", e, re, ce)
     return
})
fmt.Printf("TCF returned: %v\n", err)

err = TryCatch(func() error {
     fmt.Printf("in try\n")
     return nil
}, func(e error) (re, ce error) {
     fmt.Printf("in catch %v: %v %v\n", e, re, ce)
     return
})
fmt.Printf("TCF returned: %v\n", err)

Listing 7-5Try/Catch Emulation Example (Part 2)

```

这将输出以下内容:

```go
in try
in catch forced panic: <nil> <nil>
in finally
TCF returned: TryCatchError[forced panic <nil>]
in try
in finally
TCF returned: TryCatchError[try error <nil>]
in try
in catch forced panic: <nil> <nil>
TCF returned: TryCatchError[forced panic <nil>]
in try
TCF returned: <nil>

```

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

[https://github . com/juju/errors](https://github.com/juju/errors)

  [2](#Fn2_source)

[https://golang . org/pkg/errors/](https://golang.org/pkg/errors/)

 </aside>