# 九、接口的应用

在这一章中，我们将讨论 Java 中一些有趣的接口应用，以及它们与 Go 编码的关系。

## 界面是关键

就像在 Java 中一样，使用接口(通过具体类型)作为参数和返回类型在 Go 中很重要。它支持许多选项，比如用模拟对象 <sup>[1](#Fn1)</sup> 代替普通对象，这对于测试来说至关重要。所以，特别是当你把一个结构类型传入或传出一个函数时，看看你是否能用一个接口类型替换这个结构。如果您的函数只使用结构的方法而不使用其字段，这通常是可能的。

如果不存在与您使用的方法匹配的现有接口，请创建一个并发布给其他人使用。例如，给定这种类型:

```go
type Xxx struct {
     :
}
func (x *Xxx) DoSomethingGood() {
     :
}
func (x *Xxx) DoSomethingBad() (err error) {
     :
}

```

您可以创建接口:

```go
type DoGooder interface {
     DoSomethingGood()
}
type DoBader interface {
     DoSomethingBad() error
}

```

然后在某个使用`Xxx`的客户端，这样说:

```go
func DoWork(xxx *Xxx) {
     xxx.DoSomethingGood()
}

```

你可以把它转换成

```go
func DoWork(dg DoGooder) {
     dg.DoSomethingGood()
}

```

但是现在`DoWork`调用者可以发送一个`Xxx`的实例或者任何其他有`DoSomethingGood()`方法的类型。有时，您需要调用 struct 类型的多个方法。有两个主要选项:

1.  给函数多个参数，每个参数对应一个不同的接口类型，调用者为所有参数传递相同的对象。

2.  创建组合接口并传入该类型。

选项二通常比选项一更受欢迎。

对于选项一，这可以用作

```go
func DoWork(dg DoGooder, db DoBader) {
     dg.DoSomethingGood()
     db.DoSomethingBad()
}

```

可以这样称呼:

```go
var xxx *Xxx
:
DoWork(xxx, xxx)

```

对于选项二，组合接口可以是

```go
type DoGoodAndBad interface {
     DoGooder
     DoBader
}

```

它可以这样使用:

```go
func DoWork(dgb DoGoodAndBad) {
     dgb.DoSomethingGood()
     dgb.DoSomethingBad()
}

```

可以这样称呼:

```go
var xxx *Xxx
:
DoWork(xxx)

```

有点令人惊讶的是，这也可以这样调用(使用一个对象，而不是指向该对象的指针):

```go
var xxx Xxx
:
DoWork(xxx)

```

Go 编译器检测对象和对象指针的使用，并做正确的事情。只有接口类型的参数才会出现这种情况。

类似地，对于当前返回结构类型的函数，可以将其更改为返回多个接口或返回一个组合接口。

接口有一个问题，可能会很成问题。由于 Go 不允许相同类型的重载(相同的名称，不同的签名)函数，所以您可以很容易地用相同的方法名称创建多个接口，通常使用不同的参数和/或返回类型。但是你不能把它们组合成一个新的界面。这也意味着一个类型不能同时实现这些不同的接口。

对此没有简单的解决办法。因此，在选择接口中的方法名称时要小心，因为您最终可能会为行为保留该名称。例如，`io.Writer`接口声称`Write`方法(及其特定的参数)仅仅意味着它认为它意味着什么。在不与该接口冲突的情况下，其他接口无法为其他目的创建名为`Write`的方法。

例如，您可以创建这样的接口:

```go
type MyWriter interface {
     // vs. io.Writer: Write([]byte) (int,error)
     Write([]byte, int) error
}

```

不可能创建同时实现了`MyWriter`和`io.Writer`接口的类型。

避免这个问题的一个方法是用更长的，通常是多个单词的名字来创建方法，而把短名字留给 Go 运行时开发人员使用。

## 依赖注入研究

进一步考虑接口的使用，应该尽可能利用*依赖注入* <sup>[2](#Fn2)</sup> (DI)。DI 是一种设计方法，在这种方法中，代码被提供了它的依赖项，而不是为它自己获得它们(换句话说，让别人为你提供你所有的依赖项)。DI 将创建依赖关系的责任从依赖它的代码中分离出来。DI 实现通常要求注入的类型符合某种接口类型。

这种方法提供了更大的灵活性，尤其是在(1)测试代码(可以注入模拟对象)或者(2)配置对象之间的复杂关系时。第二种情况在 Java 开发中非常普遍，以至于一个主要的框架， *Spring* 、<sup>、 [3](#Fn3) 、T5*Spring Boot*、 <sup>[4](#Fn4)</sup> 就是为了提供这种情况而创建的。其他选项也存在，比如谷歌的 Guice。 <sup>[5](#Fn5)</sup></sup>

Wikipedia 对 DI 的定义如下:“*依赖注入将客户端依赖的创建与客户端的行为分开，这允许程序设计松散耦合，并遵循依赖倒置* <sup>[6](#Fn6)</sup> *和单一责任* <sup>[7](#Fn7)</sup> *原则*。”

Wikipedia 描述 Spring DI:*Spring 框架的核心是它的反转控制* <sup>[8](#Fn8)</sup> *(IoC)容器，它提供了使用反射配置和管理 Java 对象的一致方法。容器负责管理特定对象的对象生命周期:创建这些对象，调用它们的初始化方法，并通过将它们连接在一起来配置这些对象。由容器创建的对象也称为托管对象或 beans。对象可以通过依赖关系查找或依赖关系注入的方式获得。*

那么，什么是依赖呢？它是(至少)一个对象

1.  具有状态和/或行为。

2.  状态应该被封装(对任何用户隐藏)，以便实现可以改变。因此，行为最好用接口来表示。

3.  由一些(相关的)代码使用。

在 Spring 的例子中，有阿迪*容器*，它管理所谓的*bean*(可以链接在一起的 POJOs <sup>[9](#Fn9)</sup> )。容器通常像地图一样，提供可以在运行时解析的命名对象。在大多数情况下，容器基于工厂方法的注释(比如说`@Bean`)或外部定义(比如说 XML)来创建 bean 实例。DI 通常通过注释(比如`@Inject`或`@Wired`)来告诉容器将一个源 POJO 链接(注入)到一个目标 POJO。

容器拥有订购先决条件 bean 创建和注入。通常，beans 是单例对象(在应用程序中共享一个实例)。容器通常不是程序执行期间进出的对象的来源。通常，容器扮演主程序的角色，创建 beans，然后在程序启动时将它们“连接”在一起。

Java DI 框架经常使用反射来创建要注入的对象。他们通常采用应用程序开发人员定义的 POJO，并将其包装在一个添加额外功能的*代理* <sup>[10](#Fn10)</sup> 中，比如日志记录或数据库事务管理。代理概念的关键是，代理的客户端不能仅通过被代理的接口来区分它和它所代理的对象；它完全实现了被代理对象的行为契约，因此它是对象的直接替代。在大多数情况下，POJO 类必须实现一个或多个接口，这些接口可以具有在运行时动态定义的具体实现。

Go 目前不支持这种代理的动态创建，因为似乎不可能在运行时通过使用反射来定义类型，这可以用于实现符合接口的对象。这是经常使用代码生成方法的部分原因。或许这在未来会有所改变。Go 确实支持创建客户端可能知道的类似代理的外观对象。

让我们将术语 POGO 定义为 POJO 的 GO 等价物。POGOs 通常被实现为 Go 结构。

Go 没有标准的 DI 容器实现。Go 社区提供了一些，比如优步的 *Dig* <sup>[11](#Fn11)</sup> (或者 *Fx* <sup>[12](#Fn12)</sup> )和谷歌的 *Wire* 。 <sup>[13](#Fn13)</sup>

Dig 描述如下:

*一个基于反射的依赖注入工具包，适合于*:

*   *为应用框架提供动力*

*   *进程启动时解析对象图*

Wire 描述如下:*使用* [*依赖注入*](https://en.wikipedia.org/wiki/Dependency_injection) *自动化连接组件的代码生成工具。组件之间的依赖关系在 Wire 中表示为函数参数，鼓励显式初始化而不是全局变量。因为 Wire 在没有运行时状态或反射的情况下运行，所以编写用于 Wire 的代码即使对于手写的初始化也是有用的。*

这两个示例容器举例说明了实现 Go 容器的主要方法:

1.  使用反射(像 Spring 一样)来设置 POGOs 中的字段，以将它们连接在一起。

2.  使用代码生成来创建逻辑(很像在`main`中手动完成的那样)以将弹簧连接在一起。

DI 容器特别适合于提供依赖项，比如日志记录器、数据库连接池、数据缓存、HTTP 客户端和类似的伪全局值。事实上，如果做到极致，容器本身就是应用程序中唯一的公共顶级对象；所有其他的都由容器管理。

在 Go 中，有几个注射选项:

1.  实例初始化——在这里，依赖关系是通过在声明实例文字时设置来注入的。

2.  constructor/factory——这里，依赖关系是通过传递给一个构造函数(`New...`)或其他工厂方法来注入的。通常，这是首选选项。

3.  直接字段分配–这里，通过直接分配字段来注入依赖性。通常，该字段必须是公共的(因为依赖类型通常在不同的包中)才能实现这一点。应该避免这种选择。

4.  Setter 方法——这里，依赖项是通过传递给“setter”方法来注入的。很少这样做，因为结构通常不为所有私有字段提供 get/set 方法，尤其是作为依赖项的公共接口的一部分。

前两种形式的局限性在于，不可能建立相互循环依赖的 POGOs。一般来说，最好避免这种依赖图；依赖关系应该形成一个层次结构。对于后两种情况，依赖关系是在创建实例之后设置的，因此在没有设置依赖关系时会有一个窗口。

作为手动 DI 的一个例子，考虑清单 [9-1](#PC12) 中显示的这三种依赖类型(`Cache`、`HTTPClient`和`Logger`)。图 [9-1](#Fig1) 中通过浏览器显示的基本功能(无 DI)示例。

![img/516433_1_En_9_Fig1_HTML.jpg](img/516433_1_En_9_Fig1_HTML.jpg)

图 9-1

使用建议调用请求

```go
package main

import (
     "fmt"
     "time"
)

type Cache interface {
     Get(name string) (interface{}, bool)
     Set(name string, value interface{}) error
     ClearName(name string)
     ClearAll()
}

type MapCache map[string]interface{}

func (c MapCache) Get(name string) (res interface{}, ok bool) {
     res, ok = c[name]
     return
}
func (c MapCache) Set(name string, value interface{}) (err error) {
     c[name] = value
     return
}
func (c MapCache) ClearName(name string) {
     delete(c, name)
     return
}
func (c MapCache) ClearAll() {
     for k, _ := range c {
          delete(c, k)
     }
     return
}

type HTTPClient interface {
     SendReceive(url, method string, in interface{}) (out interface{},
          err error)
}

type EchoHTTPClient struct {
}

func (c *EchoHTTPClient) SendReceive(url, method string, in interface{}) (out interface{},
     err error) {
     out = fmt.Sprintf("SENT %s %s with %v", method, url, in)
     return
}

type Logger interface {
     Log(format string, args ...interface{})
}

type StdoutLogger struct {
}

func (l *StdoutLogger) Log(format string, args ...interface{}) {
     fmt.Printf("%s - %s\n", time.Now().Format(time.StampMilli), fmt.Sprintf(format, args...))
}

type HTTPService struct { // also a HTTPClient
     log    Logger
     client HTTPClient
     cache  Cache
     // :  other fields not using dependencies
}

func NewService(client HTTPClient, log Logger,
     cache Cache) (s *HTTPService) {
     s = &HTTPService{}
     s.log = log
     s.client = client
     s.cache = cache
     // : set other fields
     return
}

func (s *HTTPService) SendReceive(url, method string,
     in interface{}) (out interface{}, err error) {
     key := fmt.Sprintf("%s:%s", method, url)
     if xout, ok := s.cache.Get(key); ok {
          out = xout
          return
     }
     out, err = s.client.SendReceive(url, method, in)
     s.log.Log("SendReceive(%s, %s, %v)=%v", method, url, in, err)
     if err != nil {
          return
     }
     err = s.cache.Set(key, out)
     return
}

func main() {
     log := StdoutLogger{}      // concrete type
     client := EchoHTTPClient{} // concrete type
     cache := MapCache{}        // concrete type
     // create a service with all dependencies injected
     s := NewService(&client, &log, cache)
     // :
     for i:= 0; i < 5; i++ {
          if i % 3 == 0 {
               cache.ClearAll()
          }
          data, err := s.SendReceive("some URL", "GET",
             fmt.Sprintf("index=%d", i))
         if err != nil {
             fmt.Printf("Failed: %v\n", err)
           continue
          }
          fmt.Printf("Received: %v\n", data)
     }
     // :
}

Listing 9-1Dependency Injection in a Go Example

```

前面的例子展示了如何定义三个可注入的接口，并为每个接口提供了一个简单的示例(可能称为模拟)实现，然后注入每个实现。这里，`main()`函数发送五个事务，并在序列中途清空缓存。注意以下输出显示了缓存的影响(五个事务中只有两个被执行):

```go
Jul 20 09:10:40.348 - SendReceive(GET, some URL, index=0)=<nil>
Received: SENT GET some URL with index=0
Received: SENT GET some URL with index=0
Received: SENT GET some URL with index=0
Jul 20 09:10:40.349 - SendReceive(GET, some URL, index=3)=<nil>
Received: SENT GET some URL with index=3
Received: SENT GET some URL with index=3

```

Go 社区中的一些人认为使用 DI，尤其是当由容器管理时，对于 Go 来说并不习惯。通过容器的 DI 可以隐藏对象之间的关系，而在代码中手动创建它们(如前所示)则更加明显。这一论点有可取之处。但是，随着应用程序复杂性的增长和涉及的部件(POGOs)的增加，手动代码可能会失去控制，自动化 DI 解决方案可能是合适的(甚至是必要的)。

不管你如何结束这场争论，在作者看来，让你的代码能够支持 DI 是更好的方法。此外，如果谷歌和脸书都提供库做 DI，它一定是有用的。

## 关于面向方面编程

Java 支持一种叫做*面向方面编程* <sup>[14](#Fn14)</sup> (AOP)的编程风格。AOP 允许用新的行为(也就是代码)来扩充(用所谓的*建议*)代码(通常称为基础或原始代码)。维基百科是这样描述的:*一种编程范式，旨在通过允许分离横切关注点*<sup>[15](#Fn15)</sup>【XCC】*来增加模块化。它通过向现有代码添加额外的行为(一个建议)来实现这一点，而不修改代码本身，而是通过“切入点”规范单独指定修改哪个代码，例如“当函数名以‘set’开头时，记录所有函数调用”。这允许将对业务逻辑不重要的行为添加到程序中，而不会弄乱作为功能核心的代码。*

AOP 中有三个关键概念:

1.  切入点——指定在哪里应用建议；通常，一些谓词(通常是一种模式，如正则表达式)选择要建议的代码或数据。切入点通常仅限于匹配一个或多个类型中的一个或多个方法，但是一些 AOP 系统也允许匹配数据字段。许多关节点可以匹配一个切入点。

2.  建议——当切入点被触发时该做什么。建议有很多种，但最常见的是之前、之后和周围。

3.  连接点——代码中应用建议的实际位置。

切入点和建议代码通常由一个类似类的构造定义，称为*方面*，这是一种描述切入点和/或所需建议的方法。使用 Java，有几种方法可以在连接点应用建议:

1.  静态重写源代码——一些预处理器(在编译之前)编辑基本源代码。

2.  静态重写目标代码——一些后处理器(编译后)编辑基本目标代码(这在 Go 中很难做到；如果在需要编译器更改的代码生成阶段完成，会更容易)。

3.  动态重写目标代码——一些运行时处理器编辑目标代码，通常是在第一次加载时(这在 Go 中很难做到)。

4.  使用动态代理——一些运行时处理器包装代码，通常是在第一次加载时(这在 Go 中很难做到)。

Java 有几种 AOP 实现。最受欢迎的是*AspectJ*<sup>[16](#Fn16)</sup>和 *Spring* *AOP* 。 <sup>[17](#Fn17)</sup> AspectJ 更全面，主要使用增强选项二和三。Spring AOP 主要使用增强选项四。

AOP 通常用于向代码中添加行为。常见的例子是向 web API 处理程序添加日志记录、授权检查和事务支持。这些是 XCC 的例子，它们通常不是代码的主线目的或核心关注点的一部分，但是支持上下文的需要。如果主代码中没有杂乱的代码来提供它们就更好了。

Go AOP 选项有限。标准库中没有直接支持。一些社区提供的选择是存在的，但是它们可能还不成熟。它们不像 Java 产品那样全面。目前，似乎没有任何 Go AOP 产品像 Java AOP 那样支持非侵入式地(既不改变客户机也不改变服务代码)向基类型添加通知。

AOP 风格的编程看起来很“神奇”(代码有新的行为，而行为的来源并不总是很明显)。像 DI 容器一样，AOP 风格的编程在 Go 中并不习惯。但是和 DI 一样，它也是增加支持的有力手段。

在 Go 中，类似 AOP 的行为可以通过应用代码来实现，通常称为*中间件* <sup>[18](#Fn18)</sup> (又名软件胶水)。这是通过将服务包装在符合服务原型的处理器中，在客户端和服务之间添加的功能(因此称为中间功能)。由于 Go 支持一流的功能，中间件可以相对容易地实现。

注意任何 HTTP 处理程序都必须符合在`net/http`中定义的接口:

```go
type HandlerFunc func(http.ResponseWriter, *http.Request)

```

给定这些助手函数，如清单 [9-2](#PC15) 和 [9-3](#PC17) 所示(又名中间件或 around advice):

```go
package main

import (
     "fmt"
     "log"
     "net/http"
     "time"
)

func LogWrapper(f http.HandlerFunc) http.HandlerFunc {
     return func(w http.ResponseWriter, req *http.Request) {
          method, path := req.Method, req.URL
          fmt.Printf("entered handler for %s %s\n", method, path)
          f(w, req)
          fmt.Printf("exited handler for %s %s\n", method, path)
     }
}

func ElapsedTimeWrapper(f http.HandlerFunc) http.HandlerFunc {
     return func(w http.ResponseWriter, req *http.Request) {
          method, path := req.Method, req.URL
          start := time.Now().UnixNano()
          f(w, req)
          fmt.Printf("elapsed time for %s %s: %dns\n",
               method, path, time.Now().UnixNano() - start)
     }
}

Listing 9-2Advice/Middleware for HTTP Requests (Part 1)

```

请注意，这些包装器函数返回调用目标服务时应用的其他函数，而不是调用包装器时应用的函数。这两种方法都是 Around advice(最常见的一种)的示例，因为它们在调用目标服务之前和服务返回之后都采取行动。

如果您需要 around 行为来避免可能出现的混乱，请像这样重写包装器:

```go
:
defer func(){
     if p := recover(); p != nil {
          fmt.Printf("elapsed time for %s %s failed: %v\n",
               method, path, p)
          panic(p)
     }
}()
f(w, req)
:

```

例如，让我们看一下向 HTTP 请求处理程序添加日志记录和计时。

```go
var spec = ":8086"  // localhost

func main() {
     // regular HTTP request handler
     handler := func(w http.ResponseWriter, req *http.Request) {
          fmt.Printf("in handler %v %v\n", req.Method, req.URL)
          time.Sleep(1 * time.Second)
          w.Write([]byte(fmt.Sprintf("In handler for %s %s", req.Method, req.URL)))
     }
     // advised handler
     http.HandleFunc("/test", LogWrapper(ElapsedTimeWrapper(handler)))
     if err := http.ListenAndServe(spec, nil); err != nil {
          log.Fatalf("Failed to start server on %s: %v", spec, err)
     }
}

Listing 9-3Advice/Middleware for HTTP Requests (Part 2)

```

运行者:

它生成以下日志输出:

```go
entered handler for GET /test
in handler GET /test
elapsed time for GET /test: 1000141900ns
exited handler for GET /test

```

在这里，不同的中间件增加了对日志和计时的关注；原始处理程序不会受到任何影响。HTTP 引擎也不是。可以应用任意数量的包装器(以增加一些执行时间为代价)。一个成熟的 AOP 系统可能会自动应用这样的中间件，但是也可以像前面显示的那样手动应用。

<aside aria-label="Footnotes" class="FootnoteSection" epub:type="footnotes">Footnotes [1](#Fn1_source)

[T2`https://en.wikipedia.org/wiki/Mock_object`](https://en.wikipedia.org/wiki/Mock_object)

  [2](#Fn2_source)

[T2`https://en.wikipedia.org/wiki/Dependency_injection`](https://en.wikipedia.org/wiki/Dependency_injection)

  [3](#Fn3_source)

[`https://en.wikipedia.org/wiki/Spring_Framework`](https://en.wikipedia.org/wiki/Spring_Framework)；[`https://spring.io/`T5】](https://spring.io/)

  [4](#Fn4_source)

[T2`https://spring.io/projects/spring-boot`](https://spring.io/projects/spring-boot)

  [5](#Fn5_source)

[T2`https://en.wikipedia.org/wiki/Google_Guice`](https://en.wikipedia.org/wiki/Google_Guice)

  [6](#Fn6_source)

[T2`https://en.wikipedia.org/wiki/Single-responsibility_principle`](https://en.wikipedia.org/wiki/Single-responsibility_principle)

  [7](#Fn7_source)

[T2`https://en.wikipedia.org/wiki/Dependency_inversion_principle`](https://en.wikipedia.org/wiki/Dependency_inversion_principle)

  [8](#Fn8_source)

[T2`https://en.wikipedia.org/wiki/Inversion_of_control`](https://en.wikipedia.org/wiki/Inversion_of_control)

  [9](#Fn9_source)

[T2`https://en.wikipedia.org/wiki/Plain_old_Java_object`](https://en.wikipedia.org/wiki/Plain_old_Java_object)

  [10](#Fn10_source)

[`https://en.wikipedia.org/wiki/Proxy_pattern`](https://en.wikipedia.org/wiki/Proxy_pattern) ， [`https://docs.oracle.com/javase/8/docs/api/java/lang/reflect/Proxy.html`](https://docs.oracle.com/javase/8/docs/api/java/lang/reflect/Proxy.html)

  [11](#Fn11_source)

[T2`https://github.com/uber-go/dig`](https://github.com/uber-go/dig)

  [12](#Fn12_source)

[T2`https://pkg.go.dev/go.uber.org/fx`](https://pkg.go.dev/go.uber.org/fx)

  [13](#Fn13_source)

[T2`https://github.com/google/wire`](https://github.com/google/wire)

  [14](#Fn14_source)

[T2`https://en.wikipedia.org/wiki/Aspect-oriented_programming`](https://en.wikipedia.org/wiki/Aspect-oriented_programming)

  [15](#Fn15_source)

[T2`https://en.wikipedia.org/wiki/Cross-cutting_concern`](https://en.wikipedia.org/wiki/Cross-cutting_concern)

  [16](#Fn16_source)

[T2`www.eclipse.org/aspectj/`](http://www.eclipse.org/aspectj/)

  [17](#Fn17_source)

[T2`https://howtodoinjava.com/spring-aop-tutorial/`](https://howtodoinjava.com/spring-aop-tutorial/)

  [18](#Fn18_source)

[T2`https://en.wikipedia.org/wiki/Middleware`](https://en.wikipedia.org/wiki/Middleware)

 </aside>