# 十四、Go 包调查

本章包含几个 Go 软件包的简要介绍。接下来的其他章节将更详细地讨论选择包。

## 使用 Go 访问文件

Go 和 Java 有所不同的一个常用领域是访问文件和目录。Java 的设计比 Go 更倾向于将文件访问抽象化。Java 提供了多种访问模式:

1.  字节/字符流-文件看起来像字节或字符序列。

2.  字节/字符通道–支持块模式访问；面向随机和异步访问；可以跨进程共享内存。

3.  高级文件操作抽象，如复制文件或遍历目录树。

Go 提供了不太抽象但通常等价的访问，特别是对于流风格的操作。Go access 类似于 Unix 风格的文件处理。存在与 Unix 文件 API 非常匹配的 API。文件几乎总是被视为字节序列。Go 还提供了与 Unix APIs 几乎完全匹配的低级 API。

这些差异体现在 Java 和 Go APIs 的风格上。

例如，使用基本的 Java 流，可以复制这样一个文件:

```go
public static long copyFile(String fromPath, String toPath)
    throws IOException {
  try (var bis = new BufferedInputStream(new FileInputStream(
     new File(fromPath)))) {
    try (var bos = new BufferedOutputStream(new FileOutputStream(
        new File(toPath)))) {
      return copyContent(bis, bos);
    }
  }
}

```

注意两个 try 语句可以合并成一个 try 语句:

```go
public static long copyFile(String fromPath, String toPath)
    throws IOException {
  try (var bis = new BufferedInputStream(new FileInputStream(
        new File(fromPath)));
      var bos = new BufferedOutputStream(new FileOutputStream(
           new File(toPath)))) {
    return copyContent(bis, bos);
  }
}

```

文件数据由复制

```go
private static long copyContent(InputStream is, OutputStream os)
    throws IOException {
  var total = 0L;
  var buf = new byte[N]; // N at least several KB
  for (;;) {
    var count = is.read(buf, 0, buf.length);
    if (count == 0)
      break;
    os.write(buf);
    total +=buf.length;
  }
  return total;
}

```

或者更简洁地说，通过使用 JRE 库等价物:

```go
private static long copyContent(InputStream is, OutputStream os)
    throws IOException {
  return is.transferTo(os);
}

```

在 Go 中，这可能是

```go
func CopyFile(fromPath, toPath string) (count int64, err error) {
      var from, to *os.File
      if from, err = os.Open(fromPath); err != nil {
            return
      }
      defer from.Close()
      if to, err = os.Create(toPath)
            err != nil {
            return
      }
      defer to.Close()
      count, err = io.Copy(to, from)
      return
}

```

因此，我们看到 Go 倾向于更直接地使用文件类型。这是因为`File`类型实现了`io.Reader`和`io.Writer`接口。我们还看到了如何用多个`defer`语句替换`try`语句。

Go 和 Java 都可以处理目录。例如，要输出从某个根目录开始的所有文件及其大小，在 Java 中可以这样做(使用 Java 流):

```go
public static void PrintAllNames(String path) throws IOException {
  try (var walk = Files.walk(Paths.get(path))) {
    walk.filter(Files::isRegularFile).
        map(p -> String.format("%s %d", p, p.toFile().length())).
        forEach(System.out::println);
  }
}

```

这样叫:

```go
PrintAllNames(".")

```

在 Go 中，这可能是

```go
var printName = func(path string, info os.FileInfo, xerr error) error{
      if xerr != nil {  // exit fast if entered with an error
            return xerr
      }
      if info.Mode().IsRegular() {
            fmt.Println(path, info.Size())
      }
      return nil
}
func PrintAllNames(path string) (err error) {
      err = filepath.Walk(path, printName)
      return
}

```

这样叫:

```go
PrintAllNames(".")

```

Java 代码使用 Java 的函数流和方法引用/lambdas(作为回调)。Go 代码使用了一个回调函数。在这两种情况下，回调只选择文件并格式化数据。

## 压缩服务

本节概述了一些用于压缩和归档的 Go 包。

### 存档包

与 Java 一样，Go 提供了读写归档文件的功能。Java 专注于 ZIP 存档格式。Go 也支持 TAR 格式。每种格式都有子程序包:

*   `tar`–读写 tar 档案

*   `zip`–读写 ZIP 档案

`archive/tar`包提供了这些类型:

*   type Format–代表支持的 tar 格式 USTAR、PAX 和 GNU 的枚举

*   type Header–表示 tar 文件中的标题

*   类型读取器–提供对 tar 的读取权限

*   类型编写器—提供对 tar 的写访问权限

标题类型有以下方法:

*   func FileInfoHeader(fi os。FileInfo，链接字符串)(*Header，error)–从文件信息创建标题

*   func (h *Header) FileInfo() os。FileInfo–从文件头获取文件信息

读取器类型有以下方法:

*   func NewReader(r io。读者)*读者–成为读者

*   func (tr *Reader) Next() (*Header，error)-前进到下一个文件

*   func(tr * Reader)Read(b[]byte)(int，error)-从文件中读取数据

编写器类型有以下方法:

*   func NewWriter(w io。作家)*作家–成为作家

*   func (tw *Writer) Close()错误–刷新并关闭文件

*   func (tw *Writer) Flush()错误–写入缓冲数据

*   func(tw * Writer)Write(b[]byte)(int，error)–带缓冲写入数据

*   func(tw * Writer)Write Header(HDR * Header)错误–写入文件头

`archive/zip`包提供了这些类型:

*   type Compressor–将编写器转换为压缩编写器/关闭器的功能

*   类型解压缩器–将阅读器转换为解压缩阅读器/关闭器的功能

*   类型文件–表示压缩文件；包装文件头

*   type file header–表示 zip 文件中的文件；有许多有用的领域

*   ReadCloser 标牌-可读取和关闭

*   类型阅读器-可以阅读

*   打字者-可以书写

文件类型有这样的方法:

*   func (f *File) Open() (io。ReadCloser，error)–打开压缩文件

FileHeader 类型有以下方法:

*   func FileInfoHeader(fi os。FileInfo) (*FileHeader，error)-创建文件头

*   func (h *FileHeader) FileInfo() os。文件关于

*   func (h *FileHeader) ModTime()时间。时间

*   func(h * file header)mode()(OS 模式)。FileMode)

*   func(h * file header)setmodtime(t time)。时间)

*   func(h * file header)set mode(OS 模式)。FileMode)

ReadCloser 类型有以下方法:

*   func OpenReader(名称字符串)(*ReadCloser，错误)–访问压缩条目

*   func (rc *ReadCloser) Close()错误–关闭压缩条目

读取器类型有这个方法:

*   func NewReader(r io。ReaderAt，size int64) (*Reader，error)-创建一个阅读器

编写器类型有以下方法:

*   func NewWriter(w io。作家)*作家–成为作家

*   func (w *Writer) Close()错误–关闭压缩条目

*   func (w *Writer) Create(名称字符串)(io。编写器，错误)–添加压缩条目

*   func(w * Writer)create header(FH * file header)(io。编写器，错误)

*   func (w *Writer) Flush()错误–刷新任何缓冲的输出

*   func (w *Writer) SetComment(注释字符串)错误

### 压缩包

Java 支持多种环境下的数据压缩。对于存档文件，通常会这样做。Go 使压缩成为一个更加独立和通用的动作。它支持多种形式的压缩。Go 支持(通过包)这些形式的压缩:

*   `bzip2`–bzip2 解压缩

*   `gzip`–gzip 压缩数据的读写

*   `zlib`–zlib 压缩数据的读写

*   `flate`–放气压缩

*   伦佩尔-齐夫-韦尔奇压缩公司

这些可用于压缩/解压缩字节流，通常是压缩/解压缩到文件中。它们由归档包使用。有关更多详细信息，请参见 Go 包文档。

作为使用这些包(和`os`包)的例子，让我们将一个文件压缩成另一个 GZ 格式的文件。下面是一个可能的实现:

```go
func CompressFileToNewGZIPFile(path string) (err error) {
      var inFile, gzFile *os.File
      // access input file
      if inFile, err = os.Open(path); err != nil {
            return
      }
      defer inFile.Close()
      // create output file
      if gzFile, err = os.Create(path + ".gz"); err != nil {
            return
      }
      defer gzFile.Close()
      // copy input to output, compressing as copied
      w := gzip.NewWriter(gzFile)
      defer w.Close()
      _, err = io.Copy(w, inFile)
      return
}

```

请注意，如果此函数返回一个错误，则可能会用无效数据创建输出文件。

如果你替换

```go
w := gzip.NewWriter(gzFile)

```

随着

```go
w := gzFile

```

并且去掉最后一个`defer`语句，你会得到一个未压缩的文件副本。

## 图像

`image`包及其子包为读取、绘制和格式化图像提供了支持。`image`包包含使用各种形式/大小的颜色的图像表示，例如 Alpha、Alpha16、CMYK、Gray、Gray16、NRGBA、NRGBA64、NYCbCrA、Paletted、RGBA、YCbCr。它还支持几种图像类型:image、PalatedImage 和 Uniform。它还有一些关键的图像相关类型，如点和矩形。

有关使用此软件包的一些示例，请参见顶点计划。

映像包具有以下关键接口和结构:

```go
type Image interface {
      ColorModel() color.Model
      Bounds() image.Rectangle
      At(x, y int) color.Color
}

```

包`image`有一些关键方法:

*   `func Decode(r`[`io`](https://golang.org/pkg/io/)`.Reader) (Image, string, error)`–读取图像

*   `func Encode(w`[`io`](https://golang.org/pkg/io/)`.Writer, m`[`image`](https://golang.org/pkg/image/)`.Image) error`——写一个形象

```go
type Point struct {
    X, Y int
}
type Rectangle struct {
    Min, Max Point
}

```

点和矩形具有创建、调整和比较值的方法:

*   func Pt(X，Y int)Point–创建一个点

*   功能(p 点)添加(q 点)点–添加点

*   函数(p 点)除法(k 整数)点–将 p 除以 k

*   func (p 点)Eq(q 点)bool–测试相等

*   func (p 点)In(r 矩形)bool–测试 p 是否在 r 中

*   func (p 点)Mod(r 矩形)点–p 在 r 中的模数

*   函数(p 点)乘法(k 整数)点–将 p 乘以 k

*   功能(p 点)子(q 点)点–减去一个点

*   func Rect(x0，y0，x1，y1 int)Rectangle–制作一个矩形

*   功能(r 矩形)添加(p 点)矩形–添加矩形

*   func (r 矩形)At(x，y int) color。颜色测试–获取某一点的颜色

*   func(r Rectangle)Bounds()Rectangle–获取边界

*   func(r Rectangle)Canon()Rectangle–制作 r 的规范版本

*   func (r 矩形)ColorModel()颜色。模型–获取颜色模型

*   func(r Rectangle)Dx()int–获取宽度

*   func(r Rectangle)Dy()int–获取高度

*   func(r Rectangle)Empty()bool–测试是否没有包含点

*   func(r Rectangle)Eq(s Rectangle)bool–测试相等

*   func(r Rectangle)In(s Rectangle)bool–r all In s

*   func(r Rectangle)Inset(n int)Rectangle–返回 r inset 乘 n

*   func (r 矩形)Intersect(s 矩形)Rectangle–返回最大的交集

*   函数(r 矩形)重叠(s 矩形)布尔–如果交集不为空

*   func (r 矩形)大小()点

*   函数(r 矩形)子(p 点)矩形–将 r 平移-p

*   func(r Rectangle)Union(s Rectangle)Rectangle–返回覆盖 r 和 s 的矩形

所有的图像形式至少都有这些方法(有些超出了图像接口):

*   func (p * <type>) At(x，y int) color。颜色-获取某一点的值</type>

*   func(p *<type>)Bounds()Rectangle–获取图像边界</type>

*   func (p * <type>) ColorModel()颜色。模型–获取图像颜色模型</type>

*   func(p *<type>)Opaque()bool–查看图像是否不透明(没有透明单元格)</type>

*   func (p * <type>) PixOffset(x，y int)int–获取该点在像素列表中的偏移量</type>

*   func (p * <type>) Set(x，y int，c color。颜色)–在一个点上设置值</type>

*   func (p * <type>) SetAlpha(x，y int，c color。Alpha)–仅设置一个点的 alpha</type>

*   func (p * <type>)子图像(r 矩形)图像–获取图像的子集</type>

## 输入/输出

`io`和`ioutil`包提供了基本的 I/O 操作，这些操作抽象了基本操作系统(OS)提供的动作。这个`io`包主要由接口组成，而不是实现。

Go 有几个关键接口，当它们组合在一起时，允许丰富的输入/输出(I/O)功能。在最基本的层面上，I/O，就像在 Java 中一样，通常是在字节流上完成的，有时被解释为 UTF-8 字符。支持其他字符编码。

*   读取器–可以读取一个字节或一系列字节

*   writer–可以写入一个字节或一系列字节

*   seeker–可以改变流中的读/写位置(一种随机 I/O 形式)

*   closer–可以关闭对流的访问

*   前述接口的组合

许多 Go 类型都实现了这些接口，大多数用于字节，一些用于字符。例如，Go `File`类型，它允许使用一个打开的文件实例来访问它所代表的操作系统文件的内容。

`ioutil`包提供了常见文件和目录操作的实现。

Go 和 Java 都支持读写文件(或类似文件的对象)。Java 通过字节或字符流读/写器访问来支持这一点。Java 在其 NIO(新 I/O)包中也有更高的、通常更具性能的选项。Go 的访问级别通常更低，但是 Java 的缓冲流的等价物是可用的。

Go `bufio`包在非缓冲 I/O 的基础上实现了缓冲 I/O。它提供了类似 Java 支持的各种缓冲流的支持。

`io`包提供了这些类型(主要是接口)和功能:

*   ByteReader–读取一个字节

*   bytes scanner–读取和未读取的字节

*   打字机–写入一个字节

*   接近–接近

*   limited reader——有限制的阅读器

*   piped reader——来自管道的阅读器

*   piped writer–管道的编写器

*   read closer–可以读取和关闭

*   read seeker–可以阅读和搜索

*   读写关闭——可以读、写、关闭

*   read write seeker–可以读、写、查找

*   读写——可以读、写

*   读者–可以阅读

*   reader at——在某一位置的读取受限

*   reader from–剩余读取

*   阅读符文

*   阅读和未阅读符文

*   section reader–读取一段(跨度)字节

*   探索者-可以设定一个位置

*   string writer–编写一个字符串

*   write closer–可以写，关闭

*   write seeker–可以写、查找

*   写入器–可以写入字节

*   WriterAt–可以在一个位置写入一个字节

*   writer to–可以写入有限的字节

`io`包有这个键值:

```go
var EOF = errors.New("EOF")

```

`io`包具有以下功能:

*   func Copy(dst Writer，src Reader)(写入 int64，err error)-将字节 src 复制到 dst

*   func CopyBuffer(dst Writer，src Reader，buf []byte)(写入 int64，err error)–带缓冲区复制

*   func CopyN(dst Writer，src Reader，n int64) (written int64，err error)-有限制的复制

*   func Pipe() (*PipeReader，* Pipe writer)-在管道之间复制

*   func Read 至少(r Reader，buf []byte，min int) (n int，err error)-有限制地读取

*   func ReadFull(r Reader，buf []byte) (n int，err error)–读取所有可用的

*   func WriteString(w Writer，s string) (n int，err error)-写入一个字符串

*   func LimitReader(r Reader，n int 64)Reader–创建一个最多接受 n 个字节的读取器

*   func MultiReader(阅读器...Reader)阅读器——制作一个结合所有阅读器的阅读器

*   func TeeReader(r Reader，w Writer)Reader——创建一个将 r 复制到 w 的阅读器

*   func NewSectionReader(r ReaderAt，off int64，n int 64)* section reader–创建节阅读器

`ioutil`包有这个键值:

*   `var Discard` [`io`](https://golang.org/pkg/io/) `.Writer`用来丢弃任何写入其中的东西

`ioutil`包定义了这些函数:

*   func no closer(r io。读者)io。read closer–空阅读器

*   func ReadAll(r io。reader)([]字节，错误)–读取所有剩余字节

*   func ReadDir(目录名字符串)([]os。FileInfo，error)–读取目录

*   func ReadFile(文件名字符串)([]字节，错误)-读取整个文件

*   func TempDir(dir，pattern string)(名称字符串，err 错误)-创建唯一的命名目录

*   func TempFile(目录，模式字符串)(f *os。文件，err 错误)–创建一个唯一的命名文件

*   func WriteFile(文件名字符串，数据[]字节，perm os。FileMode)错误–写入整个文件

`bufio`包定义了这些函数(提供了各种扫描仪):

*   func ScanBytes(data []byte，atEOF bool) (advance int，token []byte，err error)

*   func ScanLines(data []byte，atEOF bool) (advance int，token []byte，err error)

*   func 扫描(data []byte，atEOF bool)(前进 int，token []byte，err error)

*   func 扫描字(data []byte，atEOF bool) (advance int，token []byte，err error)

`ReadWriter`类型实现了读和写接口:

*   func NewReadWriter(r *Reader，w *Writer) *ReadWriter

`Reader`类型实现了阅读器接口:

*   func NewReader(rd io。读者)*读者

*   func NewReaderSize(rd io。阅读器，尺寸 int)*阅读器

*   func(b *读取器)Buffered() int

*   func (b *Reader) Discard(n int)(丢弃的 int，err error)

*   func(b * Reader)Peek(n int)([]字节，错误)

*   func(b * Reader)Read(p[]byte)(n int，err error)

*   func (b *Reader) ReadByte()(字节，错误)

*   func (b *Reader)读取字节(delim 字节)([]字节，错误)

*   func(b * Reader)ReadLine()(line[]byte，isPrefix bool，err error)

*   func (b *Reader) ReadRune() (r rune，size int，err error)

*   func (b *Reader) ReadSlice(delim 字节)(line[]字节，err 错误)

*   func (b *Reader) ReadString(delim 字节)(String，error)

*   func(b *阅读器)复位(r io。读者)

*   func (b *Reader) Size() int

*   func (b *Reader) UnreadByte()错误

*   func(b * Reader)un readune()错误

*   func (b *Reader) WriteTo(w io。Writer) (n int64，err error)

`Scanner`类型实现扫描仪接口:

*   func 新闻扫描器(r io。阅读器)*扫描仪

*   func (s *Scanner)缓冲区(buf []byte，max int)

*   func (s *Scanner)字节()[]字节

*   func (s *Scanner) Err()错误

*   func(s *扫描仪)Scan() bool

*   func(s *扫描仪)分割(split SplitFunc)

*   func (s *Scanner) Text()字符串

`Writer`类型实现了编写器接口:

*   func NewWriter(w io。作家)*作家

*   func NewWriterSize(w io。Writer，size int) *Writer

*   func (b *Writer) Available() int

*   func (b *Writer)缓冲的()int

*   func (b *Writer) Flush()错误

*   func (b *Writer) ReadFrom(r io。读取器)(n int64，err error)

*   func (b *Writer)复位(w io。作家)

*   func (b *Writer) Size() int

*   func (b *Writer)写入(p[]字节)(nn int，err error)

*   func (b *Writer) WriteByte(c 字节)错误

*   func(b * Writer)write rune(r rune)(size int，err error)

*   func(b * Writer)WriteString(s string)(int，error)

作为使用`os`、`bufio`和其他包的一个例子，清单 [14-1](#PC16) 和 [14-2](#PC17) 展示了一个函数，它可以计算一个单词在某个文本文件中出现的次数。

```go
func CountWordsInFile(path string) (counts map[string]int, err error) {
      var f *os.File
      if f, err = os.Open(path); err != nil {
            return
      }
      defer f.Close()
      counts, err = scan(f)
      return
}

func scan(r io.Reader) (counts map[string]int, err error) {
      counts = make(map[string]int)
      s := bufio.NewScanner(r)
      s.Split(bufio.ScanWords) // make into words
      for s.Scan() {           // true while words left
            lcw := strings.ToLower(s.Text()) // get last scanned word
            counts[lcw] = counts[lcw] + 1 // missing is zero value
      }
      err = s.Err() // notice any error
      return
}

Listing 14-1Word Count Example (Part 1)

```

这将返回文件中每个(忽略大小写)单词的计数图。一个*扫描器*的工作方式很像一个 Java *迭代器*(通过它的`hasNext`和`next`方法)。函数`ScanWords`被传递给扫描器，以确定如何解析出单词。其他几种分割方法被预定义为按字节、符号或行扫描。有状态扫描器通过`Text()`(下一个扫描的字符串)和`Err()`(任何扫描错误)方法返回结果。如果没有错误，则设置文本。扫描在第一个错误时停止。

```go
path := `...\words.txt` // point to a real file
counts, err := CountWordsInFile(path)
if err != nil {
      fmt.Printf("Count failed: %v\n", err)
      return
}

fmt.Printf("Counts for %q:\n", path)
for k, v := range counts {
      fmt.Printf("  %-20s = %v\n", k, v)
}

Listing 14-2Word Count Example (Part 2)

```

现在是时候来帮助我们的同胞了！

它产生以下内容:

```go
Counts for ".../words.txt":
  Now                  = 1
  time                 = 1
  come                 = 1
  countrymen!          = 1
  our                  = 1
  is                   = 1
  the                  = 2
  to                   = 2
  aid                  = 1
  of                   = 1

```

## 字节包

在 Go 中，类型`[]byte`经常被使用，尤其是作为输入源或输出目标(I/O)。在 Java 中，使用字节数组作为目标的情况较少；相反，使用字节数组(或字符)上的流。Go `bytes`包提供了对字节片进行 I/O 的函数。

`bytes`包提供了这些类型和功能。大多数都是不言自明的，并且与`strings`包中的函数相匹配，因为字节片段通常可以被视为 ASCII 字符串，或者更具挑战性的 UTF 8 字符:

*   func Compare(a，b []byte) int

*   func 包含(b，subslice []byte) bool

*   func 包含三洋(b[]字节，char string)bool

*   func containsruna(b[]字节，r run)bool

*   函数计数(s，sep []byte) int

*   func Equal(a，b[]字节)布尔值

*   func EqualFold(s，t[]byte)bool–case 折叠 s 和 t 后等于

*   func 字段(s[]字节)[][]字节

*   func field sfunc(s[]字节，f func(rune)bool[])字节

*   func HasPrefix（s， prefix []byte） bool

*   func HasSuffix(s，后缀[]字节)布尔

*   函数索引(s，sep []byte) int

*   func index ny(s[]字节，char string)int

*   func 索引字节(b[]字节、c 字节)int

*   func 索引 unc(s[]字节，f func(r rune) bool) int

*   func index rune(s[]字节，r rune) int

*   func Join(s[][]字节，sep[]字节)[]字节

*   func LastIndex(s，sep []byte) int

*   func lastindexany(s[]字节，char string)int

*   func 负载索引字节(带[]字节、c 字节)int

*   func lastindexfunc(s[]字节，f func(r rune) bool) int

*   func Map(映射 func(r rune) rune，s []byte) []byte

*   func 重复(b[]字节，count int)[]字节

*   func Replace(s，old，new[]字节，n int)[]字节

*   func ReplaceAll(s，旧，新[]字节)[]字节

*   func runes(s[]字节)[]rune

*   func Split(s，sep[]byte)[][]byte–在 sep 处拆分，删除 sep

*   func SplitAfter(s，sep[]byte)[][]byte–在所有 sep 之后拆分

*   func SplitAfterN(s，sep []byte，n int)[][]byte–sep 后分割受 n 限制

*   func SplitN(s，sep []byte，n int) [][]byte

*   func Title(s[]byte)[]byte–字开始到标题大小写

*   func to lower(s[]字节)[]字节

*   func to title(s[]byte)[]byte–全部为标题大小写

*   func ToUpper(s []byte) []byte

*   func ToValidUTF8(s，替换[]字节)[]字节

*   func Trim(s[]字节，割集字符串)[]字节

*   func trimfunc(s[]字节，f func(r rune)bool[]字节

*   func trim left(s[]字节，割集字符串)[]字节

*   func trimleftfunc(s[]字节，f func(r rune)bool)[]字节

*   func 前缀(s，前缀[]字节)[]字节

*   func trim right(s[]字节，割集字符串)[]字节

*   func trimrghtfunc(s[]字节，f func(r rune)bool)[]字节

*   func TrimSpace(s []byte) []byte

*   func 三字尾(s，字尾[]位元组)[]位元组

缓冲器类型提供这些功能。大多数是不言自明的。`Buffer`类型提供了一种缓冲 I/O 的方法，通常来自文件或网络:

*   func new Buffer(buf[]byte)* Buffer

*   func NewBufferString(s 字符串)*缓冲区

*   func (b *Buffer) Bytes() []byte

*   func (b *Buffer) Cap() int

*   func (b *Buffer) Grow(n int)

*   func (b *Buffer) Len() int

*   func(b * Buffer)Next(n int)[]字节

*   func (b *Buffer)读取(p[]字节)(n int，err error)

*   func (b *Buffer) ReadByte()(字节，错误)

*   func(b * Buffer)Read bytes(delim byte)(line[]byte，err error)-一直读到 delim(通常是换行)

*   func (b *Buffer) ReadFrom(r io。读取器)(n int64，err error)

*   func (b *Buffer) ReadRune() (r rune，size int，err error)

*   func (b *Buffer) ReadString(delim 字节)(行字符串，err 错误)

*   func(b * Buffer)Reset()–返回起点

*   func (b *Buffer) Truncate(n int)

*   func (b *Buffer) UnreadByte()错误

*   func(b * Buffer)un readune()错误

*   func (b *Buffer)写(p[]字节)(n int，err error)

*   func (b *Buffer)写字节(c 字节)错误

*   func(b * Buffer)write rune(r rune)(n int，err error)

*   func(b * Buffer)WriteString(s string)(n int，err error)

*   func (b *Buffer) WriteTo(w io。Writer) (n int64，err error)

读取器类型提供了这些功能。大多数是不言自明的:

*   func new Reader(b[]字节)*Reader

*   func (r *Reader) Len() int

*   func (r *Reader)读取(b[]字节)(n int，err error)

*   func (r *Reader) ReadAt(b []byte，off int64) (n int，err error)

*   func (r *Reader) ReadByte()(字节，错误)

*   func(r * Reader)ReadRune()(ch rune，size int，err error)

*   func(r *读取器)复位(b[]字节)

*   func (r *Reader) Seek(偏移量 int64，where int)(int 64，error)

*   func (r *Reader) Size() int64

*   func (r *Reader) UnreadByte()错误

*   func(r * Reader)un readune()错误

*   func (r *Reader) WriteTo(w io。Writer) (n int64，err error)

## 格式化包

`fmt`包提供了一个格式化的字符串和 I/O 函数，很像 Java 的`String.format`和*PrintStream/PrintWriter*`printf`函数。

`scanner`包提供了文本扫描和标记化。

`tabwriter`包提供了一种简单但功能较低的方式来生成表格(列化)文本输出。社区提供更高的功能支持。

`fmt`包提供了这些功能:

*   func Errorf(格式字符串，a...interface { })error–从格式化字符串中出错

*   func Fprint(w io。作家...interface{}) (n int，err error)–输出到 w

*   func Fprintf(w io。编写器，格式字符串，一个...interface{}) (n int，err error)–输出到 w

*   func Fprintln(w io。作家...interface{}) (n int，err error)–输出到 w 并添加 NL

*   函数 Fscan(r io。读者 a...interface{}) (n int，err error)–按类型扫描来自 r 的输入

*   func Fscanf(r io。读取器，格式字符串，一个...interface{}) (n int，err error)–按类型扫描来自 r 的输入

*   func Fscanln(r io。读者 a...interface{}) (n int，err error)–按类型从 r 扫描输入行

*   功能打印(a...interface{}) (n int，err error)–输出到 STDOUT

*   func Printf(格式字符串，a...interface{}) (n int，err error)–输出到 STDOUT

*   func Println(a...interface{}) (n int，err error)–输出到添加了 NL 的 STDOUT

*   功能扫描(a...interface{}) (n int，err error)–扫描来自 STDIN 的输入

*   func Scanf(格式字符串，a...interface{}) (n int，err error)–扫描来自 STDIN 的输入

*   函数扫描(a...interface{}) (n int，err error)–从 STDIN 扫描行输入

*   func Sprint(a...interface { })string–输出字符串

*   func Sprintf(格式字符串，a...interface { })string–输出字符串

*   func Sprintln(a...interface { })string–输出添加了 NL 的字符串

*   func Sscan(字符串 string，a...interface{}) (n int，err error)–从字符串中扫描输入

*   func Sscanf(字符串，格式字符串，一个...interface{}) (n int，err error)–从字符串中扫描输入

*   func Sscanln(字符串 string，a...interface{}) (n int，err error)–从字符串中扫描行输入

`fmt`包提供了这些类型:

*   类型格式化程序–一种可以自我格式化的类型

*   GoStringer 类型——这种类型可以支持根据`%#v`用细节格式化自身

*   扫描仪类型–可以自定义扫描方法的类型

*   Stringer 类型–可以将自身格式化为字符串的类型

`scanner`包提供了这些类型和功能。

`Scanner`从`io.Reader`中读取字符和标记:

*   func (s *Scanner) Init(src io。阅读器)*扫描仪–制作扫描仪

*   func(s * Scanner)Next()rune–获取下一个字符

*   func(s * Scanner)Peek()rune–检查下一个字符

*   func(s * Scanner)Pos()(Pos Position)-输出位置信息

*   func(s * Scanner)Scan()rune–获取下一个令牌

*   func(s * Scanner)token text()string–获取刚刚扫描的令牌的文本

`tabwriter`包提供了这些类型和功能。

Writer 是一个具有列对齐功能的`io.Writer`:

*   func NewWriter(输出 io。Writer，minwidth，tabwidth，padding int，padchar byte，flags uint)* Writer–创建一个编写器

*   func (b *Writer) Flush()错误

*   func (b *Writer) Init(输出 io。Writer，minwidth，tabwidth，padding int，padchar byte，flags uint)* Writer–重置一个 Writer

*   func (b *Writer)写(buf []byte) (n int，err error)

## 数据收集

与 Java 不同，Go 不太依赖一组标准的*集合*(列表、地图、集合等)。)类型和关联的实现。`slice`和`map`类型满足了大部分需求。但是 Go 包含了一些特殊用途的容器库。有关更多详细信息，请参见在线软件包文档:

*   `heap`为任何堆接口实现者提供操作；仅接口。

*   `list`提供了一个双向链表。

*   `ring`提供循环列表。

包堆提供了一些实现类型必须提供的类型和函数:

*   func Fix(h Interface，I int)–元素值更改后修复

*   功能初始化(h 接口)–初始化

*   func Pop(h Interface)Interface { }–获取最低值

*   func Push(h 接口，x 接口{ })–添加新值

*   func Remove(h Interface，I int)Interface { }–删除第 I 个值

包`list`提供了这些类型和功能。

元素是列表成员:

*   func(e * Element)Next()* Element–获取任何前任

*   func(e * Element)Prev()* Element–获取任何后续元素

`List`包含元素。这些方法不言自明:

*   func New()* List–制作列表

*   func(l * List)Back()*元素–向后移动

*   func(l * List)Front()*元素–获取第一个

*   func(l * List)Init()* List –{ Re }初始化(清空)列表

*   func(l * List)InsertAfter(v interface { }，mark * Element)*元素

*   func(l * List)insert before(v interface { }，mark * Element)*元素

*   func(l * List)Len()int–获取长度

*   func (l *List) MoveAfter(e，mark *Element)

*   func (l *List) MoveBefore(e，mark *Element)

*   func(l * List)move to back(e * Element)

*   func(l * List)move to front(e * Element)

*   func (l *List)推回(v interface { })*元素

*   func (l *List)推回列表(其他*List)

*   func(l * List)push front(v interface { })*元素

*   func (l *List) PushFrontList(其他*List)

*   func (l *List) Remove(e *Element)接口{}

例如，以相反的顺序输出列表中的所有元素:

```go
var l = list.New()
for _, x := range []int{1,2,3,4,5} {
      l.PushFront(x)
}
for v := l.Front(); v != nil; v = v.Next() {
      fmt.Print(v.Value)
}
fmt.Println()

```

从而产生`54321`。

包`ring`为`Ring`类型提供了这些功能。每个环元素都有一个`Value`字段。没有什么元素是特殊的:

*   func New(n int)* Ring–创建一个包含 n 个元素的环

*   func(r * Ring)Do(f func(interface { })–对每个元素运行一个函数

*   func(r * Ring)Len()int–获取长度

*   func(r * Ring)Link(s * Ring)* Ring–将 s 插入 r

*   func(r * Ring)Move(n int)* Ring–前进 n 个元素

*   func(r * Ring)Next()* Ring–前进一个元素

*   func(r * Ring)Prev()* Ring–后退一个元素

*   func(r * Ring)Unlink(n int)* Ring–将接下来的 n 个元素组成一个环

下面是一个打印整数环的示例:

```go
 N := 5
ring := ring.New(N)  // some capacity
count := ring.Len()
// set each element to square root the element index cubed
for i := 0; i < count; i++ {
      ring.Value = math.Sqrt(float64(i * i * i))
      ring = ring.Next()
}
// now output the values; now back at start
x := 0
ring.Do(func(v interface{}) {
      fmt.Printf("Root of cube %v = %v\n", x, v)
      x++
})

```

这产生了

```go
Root of cube 0 = 0
Root of cube 1 = 1
Root of cube 2 = 2.8284271247461903
Root of cube 3 = 5.196152422706632
Root of cube 4 = 8

```

## 整理

`sort`包提供了对可比较类型的片和支持特定接口的用户定义集合进行排序的方法。

`sort`包提供了这些类型和功能:

*   func float 64s(a[]float 64)-Sort[]float 64

*   func float 64 sorted(a[]float64)bool–测试[]float 64 是否已经排序

*   func ints(a[]int)-输出[]int

*   func int sorted(a[]int)bool–测试[]int 是否已经排序

*   func IsSorted(数据接口)bool–测试是否已经排序

*   func Search(n int，f func(int)bool)int–二分搜索法 0…N-1 查找 f 为真的值

*   func SearchFloat64s(a []float64，x float 64)int–二分搜索法为 x 排序的[]float64

*   func SearchInts(a []int，x int)int–二分搜索法为 x 排序的[]int

*   func SearchStrings(a []string，x string)int–x 的二分搜索法排序[]字符串

*   func Slice(slice interface{}，less func(i，j int)bool)–Sort[]？使用 less 函数

*   func slice sorted(slice interface { }，less func(i，j int)bool)bool–测试是否已经排序

*   func slice table(slice interface { }，less func(i，j int)bool)–稳定排序[]？使用 less 函数

*   函数排序(数据接口)-排序数据

*   函数稳定(数据接口)–稳定排序数据

*   函数字符串(一个[]字符串)–排序[]字符串

*   func strings aresorted(a[]string)bool–测试是否已经排序

*   func 反转(数据接口)接口–反转数据

Type *Interface* (一个接口，不是最好的名字选择)定义了用于排序的方法。类型需要实现此接口才能排序。这就像 Java 的`Comparable`接口:

*   len()int–支持集合的长度

*   Less(i，j int)bool–比较 I 和 j 处的元素

*   Swap(i，j int)–交换 I 和 j 处的元素

Float64Slice 充当实现`Interface`接口的[]float64:

*   func (p Float64Slice)搜索(x float64) int

*   func (p Float64Slice)排序()

IntSlice 充当实现`Interface`接口的[]int:

*   func (p IntSlice)搜索(x int) int

*   func (p IntSlice)排序()

StringSlice 充当实现`Interface`接口的[]字符串:

*   func (p StringSlice)搜索(x string) int

*   func (p 字符串片)排序)

例如，要对一段字符串进行排序:

```go
in :=[...]string{"32", "-1", "0", "a"}
out :=[...]string{"32", "-1", "0", "a"}
var xout = sort.StringSlice(out[:])
xout.Sort()
fmt.Printf("in:  %v\nout: %v\n", in, xout)

```

这产生了

```go
in:  [32 -1 0 a]
out: [-1 0 32 a]

```

另一个简单的特例是对一段字符串进行排序:

```go
var sortable = []string{"32", "-1", "0", "a"}
sort.Strings(sortable)
fmt.Printf("out: %v\n", sortable)

```

这产生了

```go
out: [-1 0 32 a]

```

## 上下文包

Go 对异步行为的支持与 Java 不同。Java 没有标准的方法来观察或取消这样的动作；每个库都有不同的做法。同样，在 Java 中，类似的支持是由库和/或框架提供的，比如 Spring 框架，标准库中不包括它。

在 Go 中，`context`包支持这样做，其中*上下文*类型携带超时、中断和跨本地和可能远程 API 边界和可能进程的作用域值。许多长期运行的 Go APIs(通常期望在 goroutine 中运行)接受一个上下文参数，允许它们被异步取消或在完成时通过通道通知调用者。

`context`包具有以下类型和功能:

*   func WithCancel(父上下文)(ctx 上下文，cancel cancel func)-添加一个取消函数

*   func WithDeadline(父上下文，d 时间。Time) (Context，cancel func)-添加截止日期

*   func WithTimeout(父上下文，超时时间。Duration)(上下文，cancel func)-添加超时

在哪里

*   Cancel 函数——调用此函数(由客户端)取消操作或(由自己)在操作完成时释放资源；应该被称为“CancelOrDoneFunc ”,因为这更好地反映了它的作用。

*   截止时间–设置未来中止未完成操作的时间。

*   超时–设置延迟，在此之后中止未完成的操作(替代截止时间)。

上下文保存上下文状态。它提供以下功能:

*   func Background()Context–返回用作 With…操作的第一个输入的基本上下文；最常用的。

*   func TODO()Context–返回一个 TODO(需要更多工作)上下文；用法很少。

*   func WithValue(父上下文，键，val 接口{})上下文–向上下文添加值。

上下文具有以下特殊功能:

*   done()–返回一个通道，该通道将在操作完成时接收消息

*   err()–返回操作中发生的任何错误；通常为零

清单 [14-3](#PC26) 演示了一个简单的使用上下文来取消一个无限值生成器。

```go
func generateIntValues(ctx context.Context, values chan<- int) {
      loop: for {
            v, err := genIntValue()
            if err != nil {
                  fmt.Printf("genIntValue error: %v\n", err)
                  close(values)
                  break
            }
            select {
            case values <- v: // output value
                  fmt.Printf("generateIntValues sent: %v\n", v)
            case <-ctx.Done():
                  break loop // done when something received
            }
      }
}
func genIntValue() (v int, err error) {
      test := rand.Intn(20) % 5
      if test == 0 {
            err = errors.New(fmt.Sprintf("fake some error"))
            return
      }
      v = rand.Intn(100)
      fmt.Printf("genIntValue next: %d\n", v)
      return
}

Listing 14-3Random Int Generator (Part 1)

```

通过清单 [14-4](#PC27) 中所示的代码运行。

```go
values := make(chan int, 10)
ctx, cf := context.WithTimeout(context.Background(), 5 * time.Second)
go generateIntValues(ctx, values)
for v := range values {  // get all generated
      fmt.Printf("generateIntValues received: %d\n", v)
}
cf()
fmt.Printf("generateIntValues done\n")

Listing 14-4Random Int Generator (Part 2)

```

它会产生这样的结果:

```go
genIntValue next: 87
generateIntValues sent: 87
genIntValue next: 59
generateIntValues sent: 59
genIntValue next: 18
generateIntValues sent: 18
genIntValue error: fake some error

generateIntValues received: 87
generateIntValues received: 59
generateIntValues received: 18
generateIntValues done

```

注意超时确保生成在某个时候结束；在这个例子中没有出现这种情况。注意，因为通道有许多(比如 100 个)槽，所以生成都发生在处理一个值之前。如果通道容量设置为零，处理顺序会变得更加混杂:

```go
genIntValue next: 87
generateIntValues sent: 87
genIntValue next: 59
generateIntValues received: 87
generateIntValues received: 59
generateIntValues sent: 59
genIntValue next: 18
generateIntValues sent: 18
generateIntValues received: 18
genIntValue error: fake some error
generateIntValues done

```

## 密码术、散列法和数据编码

Go 内置了重要的加密和散列函数。这包括对多种算法和功能的支持。每个都有自己的包装。本书不会详细介绍这些子包。请参见在线软件包文档。

*   `adler32`提供 Adler-32 校验和。

*   `aes`提供 AES 加密。

*   `cipher`提供标准的分组密码模式，包装较低级别的密码实现。

*   `crc32`提供 32 位循环冗余校验校验和。

*   `crc64`提供 64 位循环冗余校验校验和。

*   `crypto`提供加密常数。

*   `des`提供数据加密标准和三重数据加密算法。

*   `dsa`提供数字签名算法。

*   `ecdsa`提供椭圆曲线数字签名算法。

*   `ed25519`提供 Ed25519 签名算法。

*   `elliptic`提供了素数域上的几条标准椭圆曲线。

*   `fnv`提供 FNV 哈希函数。

*   `hash`为哈希函数提供接口。

*   `hmac`提供键控散列消息认证码。

*   `maphash`提供字节序列的哈希函数。

*   `md5`提供 MD5 哈希算法。

*   `pkix`提供 ASN.1 解析。

*   `rand`提供密码安全的随机数生成器。

*   `rc4`提供 RC4 加密。

*   `rsa`提供 RSA 加密。

*   `sha1`提供 SHA 哈希算法。

*   `sha256`提供了几种 SHA 哈希算法。

*   `sha512`提供了几种 SHA 哈希算法。

*   `subtle`为密码代码提供帮助函数。

*   `tls`提供 TLS 1.2 和 TLS 1.3。

*   `x509`提供 X.509 编码的密钥和证书。

## 编码包

`encoding`包提供了定义如何在字节级和类似字符串的表示之间转换数据的接口。有几个不同支持的子包。本书将只详细介绍其中的几个子包:

*   `ascii85`提供 ascii85 数据编码。

*   `asn1`提供 ASN.1 数据结构的解析。

*   `base32`提供 base32 编码。

*   `base64`提供 base64 编码。

*   `binary`提供数字和字节序列之间的转换。

*   `csv`读写逗号分隔值(CSV)文件。

*   管理 gob 流，一种二进制交换的形式。

*   `hex`提供十六进制编码和解码。

*   `pem`提供 PEM 数据编码。

*   `json`提供 JSON 编码和解码。

*   `xml`提供支持 XML 名称空间的 XML 解析器。

`csv`包有这些类型和功能。

读取器解析 CSV 输入:

*   func NewReader(r io。读者)*读者–成为读者

*   func(r * Reader)Read()(record[]string，err error)-读取一行

*   func(r * Reader)Read all()(records[][]string，err error)-读取所有行

Writer 生成 CSV 输出:

*   func NewWriter(w io。作家)*作家–成为作家

*   func(w * Writer)Flush()–提交写入的内容

*   func(w * Writer)Write(record[]string)错误–写入一行

*   func(w * Writer)Write all(records[][]string)错误–写入多行

`hex`包具有以下功能:

*   func NewDecoder(r io。读者)io。读者——成为读者

*   func NewEncoder(w io。作家)木卫一。作家——成为作家

*   func Decode(dst，src []byte) (int，error)-src 十六进制字符串到 dst 字节

*   函数解码(s 字符串)([]字节，错误)–src 十六进制字符串到 dst 字节

*   func decodelen(x int)int–总是 x / 2

*   func Dump(data []byte)字符串–格式化十六进制转储

*   功能转储器(带 io。作家)木卫一。write closer–格式化十六进制转储

*   func Encode(dst，src[]byte)int–src 字节到 dst 十六进制字符串

*   func EncodeToString(src []byte)字符串–src 字节到十六进制字符串

*   func 编码器(n int)-始终 n * 2

`json`包具有以下类型和功能:

*   func Compact(dst *字节。Buffer，src []byte)错误–删除无关紧要的空白

*   func html scape(dst *字节。Buffer，src[]byte)–使嵌入 HTML 变得安全

*   func 缩进(dst *字节。Buffer，src []byte，prefix，indent string)错误–缩进 JSON

*   func Marshal(v 接口{ })([]字节，错误)-基于传递的类型生成 JSON

*   func MarshalIndent(v interface{}，prefix，indent string) ([]byte，error)-根据传递的类型生成缩进的 JSON

*   func 解组(data []byte，v interface{})错误–将 JSON 解析为传递的类型

*   func Valid(data[]byte)bool–测试 JSON 字符串的有效性

解码器解码/解析 JSON 字符串。它提供以下功能:

*   func NewDecoder(r io。阅读器)*解码器–制作解码器

*   func(dec * Decoder)Decode(v interface { })error–解码下一个 JSON 值

*   func(dec * Decoder)DisallowUnknownFields()–导致未知键出错

*   func(dec * Decoder)input offset()int 64–输入文本中的位置

*   func(dec * Decoder)More()bool–测试是否有更多数据要解析

*   func (dec *Decoder) Token() (Token，error)-获取下一个令牌

编码器编码/构建 JSON 字符串。它提供以下功能:

*   func NewEncoder(w io。writer)* Encoder–制作编码器

*   func(enc * Encoder)Encode(v interface { })错误–将值格式化为 JSON

*   func (enc *Encoder) SetEscapeHTML(在 bool 上)–转义 HTML 控制字符

*   func (enc *Encoder) SetIndent(前缀，缩进字符串)-设置缩进空格

有关使用 JSON 和 XML 编码器的示例，请参见 capstone 程序。

### Unicode 编码包

`unicode`包提供了检查和操作流行编码中的 *Unicode* 字符(即，符文)的功能。这个包有常量和变量，它们定义了主要的 Unicode 字符类别，比如`Letter`、`Digit`、`Punct`、`Space`等等。

它具有以下功能。许多人通过 Unicode 分类来测试符文的类型。这就像 Java 的`Character.isXxx(...)`方法。

*   函数在(r rune，ranges...* range table)bool–测试符文的成员资格

*   func Is(rangeTab *RangeTable，r rune)bool–测试成员的 Rune

*   func isco control(r rune)bool

*   func IsDigit（r rune） bool

*   func IsGraphic(r rune) bool

*   func IsLetter（s rune） bool

*   func ispower(r rune)bool

*   func ismak(r rune)bool

*   func IsNumber（r rune） bool

*   func is of(ranges[]* range table，r rune) bool

*   func IsPrint(r rune) bool

*   func IsPunct(r rune) bool

*   func IsSpace(r rune) bool

*   func 此符号(r rune) bool

*   func isttle(r rune)bool

*   func isper(r rune)bool

*   func SimpleFold（r rune） rune

*   Func ToLower（s rune） 符文

*   func ToTitle(r rune) rune

*   func ToUpper(r rune)符文

`unicode`包有子包:

*   为 UTF-8 提供编码和解码符文。

*   `utf16`为 UTF-16 提供编码和解码符文。

`utf8`包提供了这些功能:

*   func DecodeLastRune(p[]byte)(r rune，size int)-获取最后一个符文和长度

*   func DecodeLastRuneInString(s string)(r rune，size int)-获取最后一个符文和长度

*   func DecodeRune(p []byte) (r rune，size int)-获取第一个符文和长度

*   获得第一个符文和长度

*   func EncodeRune(p []byte，r rune)int–在 UTF-8 中制作一个符文

*   func full rune(p[]byte)bool–测试 p 是否以有效的 UTF-8 符文开始

*   测试 s 是否以有效的 UTF-8 符文开始

*   func rune Count(p[]byte)int–计算 p 中的字符数

*   func rune Count instring(s string)(n int)-以 s 为单位计算符文数

*   func rune len(r rune)int–rune UTF-8 中有多少字节

*   func RuneStart(b 字节)bool–b 是有效的 RuneStart 吗

*   func Valid(p[]byte)bool–测试 p 是一个值 rune 序列

*   一个符文可以用 UTF8 表示吗

*   func valid string(s string)bool–Test s 是一个有效的符文序列

`utf16`包提供了这些功能:

*   func Decode(s[]uint 16)[]符文–转换为符文

*   符文——将一对符文转换成一个符文

*   func Encode(s[]rune)[]uint 16–从 rune 转换而来

*   func EncodeRune(r rune) (r1，R2 rune)-转换成一对

*   测试一个符文是否需要一对