# 三、结构和接口

当你写程序时，你选择的语言的类型系统是非常重要的。类型允许您以结构化的方式组织应用程序数据，这些数据可以保存在各种数据存储区中。当您编写应用程序(尤其是业务应用程序)时，您使用各种类型来组织应用程序数据，并将这些类型的值保存到持久化存储中。当你用 Go 编写应用程序时，理解它的类型系统和设计理念是很重要的。Go 提供了`int`、`uint`、`float64`、`string`、`bool`等多种内置类型。用于存储数组、切片和映射等值集合的数据结构被称为复合类型，因为它们由其他类型(内置类型和用户定义类型)组成。除了 Go 提供的内置类型之外，还可以通过与其他类型结合来创建自己的类型。本章包含 Go 中用户定义类型的配方。

Go 为其类型系统提供了简单性和实用性，因为该语言对各种语言规范都有很大的贡献。Go 的类型系统是为解决现实世界的问题而设计的，而不是过于依赖学术理论，当你为你的应用程序设计数据模型时，它避免了很多复杂性。Go 的面向对象方法不同于其他语言，如 C++、Java 和 C#。Go 在其类型系统中不支持继承，甚至没有一个`class`关键字。Go 有一个 struct 类型，如果你想比较 Go 的类型系统和其他面向对象语言的类型系统，它类似于类。Go 中的 struct 类型是类的一个轻量级版本，它遵循一种独特的设计，这种设计更倾向于组合而不是继承。

## 3-1.创建用户定义的类型

### 问题

您希望创建用户定义的类型来组织您的应用程序数据。

### 解决办法

Go 有一个 struct 类型，允许您通过与其他类型组合来创建用户定义的类型。

### 它是如何工作的

Go `struct`允许您通过组合一个或多个类型来创建自己的类型，包括内置类型和用户定义类型。结构是在 Go 中创建具体的用户定义类型的唯一方法。当您使用`struct`创建自己的类型时，重要的是要理解 Go 不支持其类型系统中的继承，但是它支持类型的组合，这允许您通过组合较小的类型来创建较大的类型。Go 的设计理念是通过组合较小的和模块化的组件来创建较大的组件。如果你是一个务实的程序员，你会欣赏 Go 的设计哲学，因为它有实际的好处，所以更喜欢组合而不是继承。类型的继承有时会在可维护性方面带来实际挑战。

#### 声明结构类型

关键字`struct`用于将类型声明为 struct。清单 [3-1](#Par7) 展示了一个表示客户信息的示例结构。

```go
type Customer struct {
       FirstName string
       LastName  string
       Email     string
       Phone     string
}
Listing 3-1.Declare Struct Type

```

声明了一个结构类型`Customer`，它有四个`string`类型的字段。注意，`Customer`结构及其字段被导出到其他包中，因为标识符是以大写字母开头的。在 Go 中，如果名称以大写字母开头，标识符将被导出到其他包中；否则，包内的可访问性将受到限制。如果一组结构字段有一个共同的类型，你可以在一个单行语句中组织相同类型的字段，如清单 [3-2](#Par9) 所示。

```go
type Customer struct {
    FirstName, LastName, Email, Phone string
}
Listing 3-2.Declare Struct Type

```

因为`Customer`结构的所有字段都有`string`类型，所以可以在一条语句中指定字段。

#### 创建结构类型的实例

您可以通过声明一个`struct`变量或使用 struct 文字来创建`struct`类型的实例。清单 [3-3](#Par12) 显示了通过声明一个`struct`变量并将值赋给 struct 的字段来创建一个`Customer` struct 实例的代码块。

```go
var c Customer
c.FirstName = "Alex"
c.LastName = "John"
c.Email = "alex@email.com"
c.Phone = "732-757-2923"
Listing 3-3.Creating a Struct Instance and Assigning Values

```

创建了一个`Customer`类型的实例，并将值逐个分配给结构字段。struct 文本也可以用于创建`struct`类型的实例。清单 [3-4](#Par14) 显示了代码块，该代码块通过使用一个结构文本并给该结构的字段赋值来创建一个`Customer`结构的实例。

```go
c := Customer{
       FirstName: "Alex",
       LastName:  "John",
       Email:     "alex@email.com",
       Phone:     "732-757-2923",
}              
Listing 3-4.Creating a Struct Instance Using a Struct Literal

```

使用 struct 文本创建一个`Customer`类型的实例，并将值赋给 struct 字段。请注意，即使在 struct 的最后一个字段初始化之后，也会添加一个逗号。当使用 struct 文本创建结构的实例时，可以将值初始化为多行语句，但即使在结构字段的赋值结束后也必须加上逗号。在清单 [3-4](#Par14) 中，您通过指定结构字段来初始化值。如果您清楚地知道字段的顺序，您可以在初始化值时省略字段标识符，如清单 [3-5](#Par16) 所示。

```go
c := Customer{
       "Alex",
       "John",
       "alex@email.com",
       "732-757-2923",
}             
Listing 3-5.Creating a Struct Instance Using a Struct Literal

```

当您使用 struct 文本创建 struct 实例时，您可以向 struct 的特定字段提供值，如清单 [3-6](#Par18) 所示。

```go
c := Customer{
       FirstName: "Alex",
       Email:     "alex@email.com",
}             
Listing 3-6.Creating a Struct Instance Using a Struct Literal by Specifying Values to a Few Fields

```

#### 使用用户定义的类型作为字段的类型

使用内置类型的字段创建了`Customer`结构。您可以使用其他结构类型作为结构字段的类型。让我们扩展一下`Customer`结构，添加一个新字段来保存地址信息，用一个结构作为新字段的类型。清单 [3-7](#Par20) 显示了通过添加一个新字段来扩展的`Customer`结构，该字段的类型是`Address`类型的一部分。

```go
type Address struct {
    Street, City, State, Zip string
    IsShippingAddress        bool
}

type Customer struct {
    FirstName, LastName, Email, Phone string
    Addresses                         []Address
}      

Listing 3-7.
Customer Struct with a Slice of a User-Defined Type as the Type for Field

```

通过添加一个新字段`Addresses`扩展了`Customer`结构，该字段的类型被指定为一个名为`Address`的结构的一部分。使用`Addresses`字段，您可以为一个客户指定多个地址。`IsShippingAddress`字段用于指定默认发货地址。清单 [3-8](#Par22) 显示了创建这个修改过的`Customer`结构的实例的代码块。

```go
c := Customer{
    FirstName: "Alex",
    LastName:  "John",
    Email:     "alex@email.com",
    Phone:     "732-757-2923",
    Addresses: []Address{
        Address{
            Street:            "1 Mission Street",
            City:              "San Francisco",
            State:             "CA",
            Zip:               "94105",
            IsShippingAddress: true,
        },
        Address{
            Street: "49 Stevenson Street",
            City:   "San Francisco",
            State:  "CA",
            Zip:    "94105",
        },
    },
}
Listing 3-8.Creating an Instance of Customer Struct

```

通过创建一个长度为两个值的`Address`类型的切片来初始化`Addresses`字段。

## 3-2.向结构类型添加方法

### 问题

您希望将行为添加到`struct`类型中，以提供对`struct`的操作，并作为方法调用。

### 解决办法

Go 的类型系统允许你使用一个方法接收器向结构类型添加方法。方法接收器指定哪种类型必须将函数作为方法关联到该类型。

### 它是如何工作的

在 Go 中，方法是一个由接收者指定的函数。让我们给`Customer`结构添加一个方法。

```go
func (c Customer) ToString() string {
       return fmt.Sprintf("Customer: %s %s, Email:%s", c.FirstName, c.LastName, c.Email)
}

```

方法`ToString`被添加到`Customer`结构中。在方法名之前使用一个额外的参数部分来指定接收方。在方法内部，您可以使用 receiver 的标识符来访问 receiver 类型的字段。`ToString`方法通过访问 struct 字段以字符串形式返回客户名称和电子邮件。

```go
return fmt.Sprintf("Customer: %s %s, Email:%s", c.FirstName, c.LastName, c.Email)

```

清单 [3-9](#Par31) 展示了一个示例程序，它声明了`Customer`结构并向其中添加了一些方法。

```go
package main

import (
       "fmt"
)

type Address struct {
       Street, City, State, Zip string
       IsShippingAddress        bool
}

type Customer struct {
       FirstName, LastName, Email, Phone string
       Addresses                         []Address
}

func (c Customer) ToString() string {
       return fmt.Sprintf("Customer: %s %s, Email:%s", c.FirstName, c.LastName, c.Email)
}
func (c Customer) ShippingAddress() string {
       for _, v := range c.Addresses {
              if v.IsShippingAddress == true {
                     return fmt.Sprintf("%s, %s, %s, Zip - %s", v.Street, v.City, v.State, v.Zip)
              }
       }
       return ""
}

func main() {
       c := Customer{
              FirstName: "Alex",
              LastName:  "John",
              Email:     "alex@email.com",
              Phone:     "732-757-2923",
              Addresses: []Address{
                     Address{
                            Street:            "1 Mission Street",
                            City:              "San Francisco",
                            State:             "CA",
                            Zip:               "94105",
                            IsShippingAddress: true,
                     },
                     Address{
                            Street: "49 Stevenson Street",
                            City:   "San Francisco",
                            State:  "CA",
                            Zip:    "94105",
                     },
              },
       }
       fmt.Println(c.ToString())
       fmt.Println(c.ShippingAddress())

}

Listing 3-9.Struct with Methods

```

通过指定方法接收器，`Customer`结构被附加到几个方法上。`ToString`返回客户姓名和电子邮件，`ShippingAddress`从存储在`Addresses`字段中的地址列表中返回默认送货地址。在`main`函数中，创建了一个`Customer`结构的实例，并调用了它的方法。

运行该程序时，您应该会看到以下输出:

```go
Customer: Alex John, Email:alex@email.com
1 Mission Street, San Francisco, CA, Zip - 94105

```

方法是带有接收器的函数。有两种类型的方法接收器:指针接收器和值接收器。清单 [3-9](#Par31) 中的程序使用一个值接收器向`Customer`结构添加方法。当用指针接收器指定方法时，用指向接收器值的指针调用该方法，当用值接收器指定方法时，使用接收器值的副本。因此，如果您想要改变接收器的状态(字段值),您必须使用指针接收器。

让我们给`Customer`结构添加一个新方法(参见清单 [3-9](#Par31) )来探索指针接收器。首先，让我们通过指定不带指针的接收者来添加方法。

```go
func (c Customer) ChangeEmail(newEmail string) {
       c.Email = newEmail
}

```

新添加的`ChangeEmail`方法为`Email`字段分配一个新的电子邮件地址。让我们创建一个`Customer`结构的实例，并通过传递一个新的电子邮件地址来调用`ChangeEmail`方法。

```go
c := Customer{
              FirstName: "Alex",
              LastName:  "John",
              Email:     "alex@gmail.com",
              Phone:     "732-757-2923",
              Addresses: []Address{
                     Address{
                            Street:            "1 Mission Street",
                            City:              "San Francisco",
                            State:             "CA",
                            Zip:               "94105",
                            IsShippingAddress: true,
                     },
                     Address{
                            Street: "49 Stevenson Street",
                            City:   "San Francisco",
                            State:  "CA",
                            Zip:    "94105",
                     },
              },
       }

       // Call ChangeEmail
               c.ChangeEmail("alex.john@gmail.com")
       fmt.Println(c.ToString())

```

运行该程序时，您应该会看到以下输出:

```go
Customer: Alex John, Email:alex@gmail.com

```

您已经向`ChangeEmail`方法提供了一个新的电子邮件来更改电子邮件地址，但是当您调用`ToString`方法时，它并没有反映出来。您仍然会从电子邮件字段收到旧电子邮件。若要修改方法内部结构值的状态，必须用指针接收器声明方法，以便字段值的更改将反映在方法外部。清单 [3-10](#Par43) 修改了`ChangeEmail`方法，用一个指针接收器来指定，这样对 Email 字段的更改将会在`ChangeEmail`方法之外得到反映。

```go
func (c *Customer) ChangeEmail(newEmail string) {
       c.Email = newEmail
}
Listing 3-10.A Method to Customer Struct with a Pointer Receiver

```

让我们创建一个`Customer` struct 的实例，并通过传递一个新的电子邮件地址来调用`ChangeEmail`方法。

```go
c := Customer{
              FirstName: "Alex",
              LastName:  "John",
              Email:     "alex@gmail.com",
              Phone:     "732-757-2923",
}

// Call ChangeEmail
 c.ChangeEmail(alex.john@gmail.com)
 fmt.Println(c.ToString())

```

运行该程序时，您应该会看到以下输出:

```go
Customer: Alex John, Email:alex.john@gmail.com
1 Mission Street, San Francisco, CA, Zip - 94105

```

输出显示`Email`字段的值已经改变。这里，`Customer`类型的值用于调用用指针接收器指定的`ChangeEmail`方法。

以下代码块使用类型为`Customer`的指针来调用通过指针接收器指定的`ChangeEmail`方法:

```go
c := $Customer{
                FirstName: "Alex",
                LastName:  "John",
                Email:     "alex@gmail.com",
                Phone:     "732-757-2923",
        }

// Call ChangeEmail
 c.ChangeEmail(alex.john@gmail.com)

```

值得注意的是，您可以向任何类型添加方法，包括内置类型。您可以向基元类型、复合类型和用户定义的类型添加方法。您可以为指针或值接收器类型定义方法，因此了解何时在方法上为接收器使用值或指针非常重要。简而言之，如果方法需要改变接收方的状态，接收方必须是指针。如果接收器是大型结构、数组或切片，指针接收器会更有效，因为它避免了在方法调用时复制大型数据结构的值。如果一个方法被指定了一个指针接收器，可能是为了改变接收器，那么最好在相同接收器类型的所有方法上使用指针接收器，这为用户提供了更好的可用性和可读性。

`Customer`结构的`ChangeEmail`方法需要改变它的接收者。因此，为了更好的可用性和清晰性，让我们修改其他方法。清单 [3-11](#Par53) 修改了清单 [3-9](#Par31) 的程序，所有方法都由指针接收器指定。

```go
package main

import (
        "fmt"
)

type Address struct {
        Street, City, State, Zip string
        IsShippingAddress        bool
}

type Customer struct {
        FirstName, LastName, Email, Phone string
        Addresses                         []Address
}

func (c *Customer) ToString() string {
        return fmt.Sprintf("Customer: %s %s, Email:%s", c.FirstName, c.LastName, c.Email)
}
func (c *Customer) ChangeEmail(newEmail string) {
        c.Email = newEmail
}
func (c *Customer) ShippingAddress() string {
        for _, v := range c.Addresses {
                if v.IsShippingAddress == true {
                        return fmt.Sprintf("%s, %s, %s, Zip - %s", v.Street, v.City, v.State, v.Zip)
                }
        }
        return ""
}

func main() {

        c := &Customer{
                FirstName: "Alex",
                LastName:  "John",
                Email:     "alex@email.com",
                Phone:     "732-757-2923",
                Addresses: []Address{
                        Address{
                                Street:            "1 Mission Street",
                                City:              "San Francisco",
                                State:             "CA",
                                Zip:               "94105",
                                IsShippingAddress: true,
                        },
                        Address{
                                Street: "49 Stevenson Street",
                                City:   "San Francisco",
                                State:  "CA",
                                Zip:    "94105",
                        },
                },
        }

        fmt.Println(c.ToString())
        c.ChangeEmail("alex.john@gmail.com")
        fmt.Println("Customer after changing the Email:")
        fmt.Println(c.ToString())
        fmt.Println(c.ShippingAddress())

}

Listing 3-11.Struct with Pointer Receiver on Methods

```

因为`ChangeEmail`方法需要改变接收器，所以所有方法都用指针接收器来定义。值得注意的是，您可以将方法与值和指针接收器混合使用。在前面的程序中，使用地址操作符(`&`)创建了一个指针`Customer`:

```go
c := &Customer{}

```

`Customer`指针`c`用于调用`Customer`结构的方法:

```go
fmt.Println(c.ToString())
c.ChangeEmail("alex.john@gmail.com")
fmt.Println(c.ToString())
fmt.Println(c.ShippingAddress())

```

运行该程序时，您应该会看到以下输出:

```go
Customer: Alex John, Email:alex@email.com
Customer after changing the Email:
Customer: Alex John, Email:alex.john@gmail.com
1 Mission Street, San Francisco, CA, Zip - 94105

```

## 3-3.使用类型嵌入合成类型

### 问题

您希望通过组合其他类型来创建类型。

### 解决办法

Go 支持将类型嵌入到其他类型中，这允许您通过组合其他类型来创建类型。

### 它是如何工作的

Go 的类型系统强化了组合优先于继承的设计理念，允许你通过嵌入其他类型来创建类型。通过使用通过类型嵌入实现的复合设计理念，您可以通过组合较小的类型来创建较大的类型。

让我们通过在类型中嵌入其他类型来创建类型。清单 [3-12](#Par64) 展示了可以用来在电子商务系统中表示订单的数据模型。

```go
type Address struct {
        Street, City, State, Zip string
        IsShippingAddress        bool
}

type Customer struct {
        FirstName, LastName, Email, Phone string
        Addresses                         []Address
}

type Order struct {

        Id int

        Customer

        PlacedOn   time.Time

        Status     string

        OrderItems []OrderItem

}

type OrderItem struct {
        Product
        Quantity int
}

type Product struct {
        Code, Name, Description string
        UnitPrice               float64
}

Listing 3-12.Data Model for Order Entity

```

在清单 [3-12](#Par64) 中，`Order`结构是通过嵌入另一种类型`Customer`结构来声明的。`Order`结构用于为客户下订单，因此`Customer`结构被嵌入到`Order`结构中。要嵌入一个类型，只需指定要嵌入到另一个类型中的类型的名称。

```go
type Order struct {
        Customer

}

```

由于类型嵌入，`Customer`结构的字段和行为在`Order`结构中可用。`Customer`结构将`Address`结构的片用于`Addresses`字段。`Order`结构将`OrderItem`结构的片用于`OrderItems`字段。`Product`结构被嵌入到`OrderItem`结构中。在这里，您通过组合几个其他结构类型来创建一个更大的类型`Order`结构。

让我们向为表示订单信息而声明的结构类型添加操作。清单 [3-13](#Par69) 显示了带有各种行为的`Order`的数据模型的完整版本。

```go
package main

import (
    "fmt"
    "time"
)

type Address struct {
    Street, City, State, Zip string
    IsShippingAddress        bool
}

type Customer struct {
    FirstName, LastName, Email, Phone string
    Addresses                         []Address
}

func (c Customer) ToString() string {
    return fmt.Sprintf("Customer: %s %s, Email:%s", c.FirstName, c.LastName, c.Email)
}
func (c Customer) ShippingAddress() string {
    for _, v := range c.Addresses {
        if v.IsShippingAddress == true {
            return fmt.Sprintf("%s, %s, %s, Zip - %s", v.Street, v.City, v.State, v.Zip)
        }
    }
    return ""
}

type Order struct {
    Id int
    Customer
    PlacedOn   time.Time
    Status     string
    OrderItems []OrderItem
}

func (o *Order) GrandTotal() float64 {
    var total float64
    for _, v := range o.OrderItems {
        total += v.Total()
    }
    return total
}
func (o *Order) ToString() string {
    var orderStr string
    orderStr = fmt.Sprintf("Order#:%d, OrderDate:%s, Status:%s, Grand Total:%f\n", o.Id, o.PlacedOn, o.Status, o.GrandTotal())
    orderStr += o.Customer.ToString()
    orderStr += fmt.Sprintf("\nOrder Items:")
    for _, v := range o.OrderItems {
        orderStr += fmt.Sprintf("\n")
        orderStr += v.ToString()
    }
    orderStr += fmt.Sprintf("\nShipping Address:")
    orderStr += o.Customer.ShippingAddress()
    return orderStr
}
func (o *Order) ChangeStatus(newStatus string) {
    o.Status = newStatus
}

type OrderItem struct {
    Product
    Quantity int
}

func (item OrderItem) Total() float64 {
    return float64(item.Quantity) * item.Product.UnitPrice
}
func (item OrderItem) ToString() string {
    itemStr := fmt.Sprintf("Code:%s, Product:%s -- %s, UnitPrice:%f, Quantity:%d, Total:%f",
        item.Product.Code, item.Product.Name, item.Product.Description, item.Product.UnitPrice, item.Quantity, item.Total())
    return itemStr

}

type Product struct {
    Code, Name, Description string
    UnitPrice               float64
}

Listing 3-13.Data Model for Order Entity with Operations in models.go

```

`Order`结构的`ToString`方法返回一个提供订单所有信息的`string`值。`ToString`调用其嵌入类型`Customer`的`ToString`和`ShippingAddress`方法。`ToString`方法还通过迭代`OrderItems`字段来调用`OrderItem`结构的`ToString`方法，该字段是`OrderItem`的一部分。

```go
orderStr += o.Customer.ToString()
    orderStr += fmt.Sprintf("\nOrder Items:")
    for _, v := range o.OrderItems {
        orderStr += fmt.Sprintf("\n")
        orderStr += v.ToString()
    }
    orderStr += fmt.Sprintf("\nShipping Address:")
    orderStr += o.Customer.ShippingAddress()

```

`Order`结构的`GrandTotal`方法返回订单的总计值，它调用`OrderItem`结构的`Total`方法来确定每个订单项的总值。

```go
func (o *Order) GrandTotal() float64 {
    var total float64
    for _, v := range o.OrderItems {
        total += v.Total()
    }
    return total
}

```

注意，`Order`结构的`ChangeStatus`方法改变了`Status`字段的状态，因此该方法使用了指针接收器。

```go
func (o *Order) ChangeStatus(newStatus string) {
    o.Status = newStatus
}

```

因为`ChangeStatus`方法需要一个指针接收器，所以`Order`结构的所有其他方法都是用指针接收器定义的。

清单 [3-14](#Par78) 显示了`main`函数，该函数用于创建`Order`结构的一个实例，并调用其`ToString`方法来获取订单信息。

```go
package main

import (
    "fmt"
    "time"
)

func main() {
    order := &Order{
        Id: 1001,
        Customer: Customer{
            FirstName: "Alex",
            LastName:  "John",
            Email:     "alex@email.com",
            Phone:     "732-757-2923",
            Addresses: []Address{
                Address{
                    Street:            "1 Mission Street",
                    City:              "San Francisco",
                    State:             "CA",
                    Zip:               "94105",
                    IsShippingAddress: true,
                },
                Address{
                    Street: "49 Stevenson Street",
                    City:   "San Francisco",
                    State:  "CA",
                    Zip:    "94105",
                },
            },
        },
        Status:   "Placed",
        PlacedOn: time.Date(2016, time.April, 10, 0, 0, 0, 0, time.UTC),
        OrderItems: []OrderItem{
            OrderItem{
                Product: Product{
                    Code:        "knd100",
                    Name:        "Kindle Voyage",
                    Description: "Kindle Voyage Wifi, 6 High-Resolution Display",
                    UnitPrice:   220,
                },
                Quantity: 1,
            },
            OrderItem{
                Product: Product{
                    Code:        "fint101",
                    Name:        "Kindle Case",
                    Description: "Fintie Kindle Voyage SmartShell Case",
                    UnitPrice:   10,
                },
                Quantity: 2,
            },
        },
    }

    fmt.Println(order.ToString())
    // Change Order status
    order.ChangeStatus("Processing")
    fmt.Println("\n")
    fmt.Println(order.ToString())
}

Listing 3-14.Entry Point of the Program That Creates an Instance of the Order struct in main.go

```

通过为字段提供值来创建`Order`结构的实例，包括嵌入的类型。这里使用了一个指针变量来调用`Order`结构的方法。`ToString`方法提供了客户所下订单的所有信息。`ChangeStatus`方法用于改变订单的状态，从而改变`Status`字段的值。嵌入类型时，可以提供类似于结构的普通字段的值。

运行该程序时，您应该会看到以下输出:

```go
Order#:1001, OrderDate:2016-04-10 00:00:00 +0000 UTC, Status:Placed, Grand Total:240.000000
Customer: Alex John, Email:alex@email.com
Order Items:
Code:knd100, Product:Kindle Voyage -- Kindle Voyage Wifi, 6 High-Resolution Display, UnitPrice:220.000000, Quantity:1, Total:220.000000
Code:fint101, Product:Kindle Case -- Fintie Kindle Voyage SmartShell Case, UnitPrice:10.000000, Quantity:2, Total:20.000000
Shipping Address:1 Mission Street, San Francisco, CA, Zip - 94105

Order#:1001, OrderDate:2016-04-10 00:00:00 +0000 UTC, Status:Processing, Grand Total:240.000000
Customer: Alex John, Email:alex@email.com
Order Items:
Code:knd100, Product:Kindle Voyage -- Kindle Voyage Wifi, 6 High-Resolution Display, UnitPrice:220.000000, Quantity:1, Total:220.000000
Code:fint101, Product:Kindle Case -- Fintie Kindle Voyage SmartShell Case, UnitPrice:10.000000, Quantity:2, Total:20.000000
Shipping Address:1 Mission Street, San Francisco, CA, Zip - 94105

```

该输出显示订单信息，包括总计，这是通过调用类型的相应方法计算的。

## 3-4.使用界面

### 问题

您希望创建一个接口类型，将其作为其他类型的协定提供。

### 解决办法

Go 有一个用户定义的接口类型，可以作为具体类型的契约。Go 的接口类型为您的 Go 应用程序提供了大量的可扩展性和可组合性。用关键字`interface`定义接口类型。

### 它是如何工作的

Go 的`interface`类型为您的 Go 应用程序提供了大量的可扩展性和可组合性。像 C#和 Java 这样的编程语言都支持接口类型，但是 Go 的`interface`类型在设计理念上是独一无二的。

#### 声明接口类型

与 C#和 Java 不同，在 Go 中，你不需要通过指定任何关键字来显式地将一个`interface`实现到一个具体的类型中。要将一个`interface`实现为一个具体的类型，只需提供与在`interface`类型中定义的相同签名的方法。清单 [3-15](#Par87) 显示了一种`interface`类型。

```go
type TeamMember interface {
    PrintName()
    PrintDetails()
}
Listing 3-15.Interface Type TeamMember

```

`interface`类型`TeamMember`是在团队中创建各种员工类型的契约。`TeamMember`接口在其契约中提供了两个行为:`PrintName`和`PrintDetails.`

#### 将接口实现为具体类型

让我们通过实现接口的两个行为`PrintName`和`PrintDetails`来创建一个具体类型的`TeamMember`接口。清单 [3-16](#Par90) 显示了一个具体的`TeamMember`类型，它实现了在`interface`类型中定义的方法。

```go
type Employee struct {
    FirstName, LastName string
    Dob                 time.Time
    JobTitle, Location  string
}

func (e Employee) PrintName() {
    fmt.Printf("\n%s %s\n", e.FirstName, e.LastName)
}

func (e Employee) PrintDetails() {
    fmt.Printf("Date of Birth: %s, Job: %s, Location: %s\n", e.Dob.String(), e.JobTitle, e.Location)
}

Listing 3-16.Concrete Type of TeamMember

```

一个 struct `Employee`用保存其状态的字段和基于在`TeamMember`接口中定义的行为实现的方法来声明。您不需要使用任何语法来将`interface`实现到类型中。相反，只需提供在接口中定义了相同签名的方法，就像您为实现`TeamMember`接口的`Employee`类型所做的那样。

一个`interface`类型的最大好处是它允许你为同一个`interface`类型创建不同的实现，这支持了更高层次的可扩展性。

清单 [3-17](#Par94) 显示了通过嵌入`Employee`类型创建的`TeamMember`接口的实现，它是`TeamMember`接口的实现。

```go
type Developer struct {
    Employee //type embedding for composition
    Skills   []string
}
Listing 3-17.Type Developer Implements TeamMember Interface

```

声明了一个结构`Developer`，其中嵌入了类型`Employee`。在这里你创建了更多具体类型的`TeamMember`接口。因为类型`Employee`是`TeamMember`接口的实现，所以类型`Developer`也是`TeamMember`接口的实现。类型`Employee`中定义的所有字段和方法在`Developer`类型中也可用。除了`Employee`的嵌入类型外，`Developer`结构还提供了一个`Skill`字段来表示`Developer`类型的技能。

清单 [3-18](#Par97) 显示了创建一个`Developer`实例并通过嵌入类型`Employee`调用可用方法的代码块。

```go
d := Developer{
                Employee{
                        "Steve",
                        "John",
                        time.Date(1990, time.February, 17, 0, 0, 0, 0, time.UTC),
                        "Software Engineer",
                        "San Francisco",
                },
                []string{"Go", "Docker", "Kubernetes"},
        }
        d.PrintName()
        d.PrintDetails()
Listing 3-18.Create an Instance of Developer Type and Call Methods

```

运行该程序时，您应该会看到以下输出:

```go
Steve John
Date of Birth: 1990-02-17 00:00:00 +0000 UTC, Job: Software Engineer, Location: San Francisco

```

输出显示了在 Employee 结构中定义的方法可以通过 Developer 结构的实例进行访问。

与`Employee`类型相比，`Developer`结构更像是`TeamMember`接口的具体实现。`Employee`类型是为类型嵌入而定义的，用于更具体地实现`TeamMember`接口，比如`Developer`结构。此时，`Developer`结构使用在`Employee`结构中定义的方法。因为`Developer`结构更像是一个具体的实现，它可能有自己的方法实现。这里的`Developer`结构可能需要覆盖`Employee`结构中定义的方法来提供额外的功能。清单 [3-19](#Par102) 显示了覆盖`Developer`结构的方法`PrintDetails`的代码块。

```go
// Overrides the PrintDetails
func (d Developer) PrintDetails() {
    // Call Employee PrintDetails
    d.Employee.PrintDetails()
    fmt.Println("Technical Skills:")
    for _, v := range d.Skills {
        fmt.Println(v)
    }
}
Listing 3-19.Overrides for the PrintDetails Method for the Developer struct

```

这里你调用了`Employee`的`PrintDetails`方法，并为`Developer`结构提供了一个额外的功能。

让我们创建另一个`struct`类型来提供`TeamMember`接口的不同实现。清单 [3-20](#Par105) 显示了一个名为`Manager`的结构，它通过嵌入`Employee`类型和覆盖`PrintDetails`方法来实现`TeamMember`接口。

```go
type Manager struct {
    Employee  //type embedding for composition
    Projects  []string
    Locations []string
}

// Overrides the PrintDetails
func (m Manager) PrintDetails() {
    // Call Employee PrintDetails
    m.Employee.PrintDetails()
    fmt.Println("Projects:")
    for _, v := range m.Projects {
        fmt.Println(v)
    }
    fmt.Println("Managing teams for the locations:")
    for _, v := range m.Locations {
        fmt.Println(v)
    }
}

Listing 3-20.Type Manager Implements the TeamMember Interface

```

除了`Employee`的嵌入类型之外，`Manager`结构还提供了`Projects`和`Locations`字段来表示经理管理的项目和位置。

到目前为止，您已经创建了一个名为`TeamMember`的接口类型，以及实现`TeamMember`接口的三个具体类型:`Employee`、`Developer`和`Manager`。让我们创建一个示例程序来探索这些类型并演示`interface`类型。清单 [3-21](#Par108) 显示了一个示例程序，它通过使用我们在本节中讨论过的类型来演示`interface`。

```go
package main

import (
    "fmt"
    "time"
)

type TeamMember interface {
    PrintName()
    PrintDetails()
}

type Employee struct {
    FirstName, LastName string
    Dob                 time.Time
    JobTitle, Location  string
}

func (e Employee) PrintName() {
    fmt.Printf("\n%s %s\n", e.FirstName, e.LastName)
}

func (e Employee) PrintDetails() {
    fmt.Printf("Date of Birth: %s, Job: %s, Location: %s\n", e.Dob.String(), e.JobTitle, e.Location)
}

type Developer struct {
    Employee //type embedding for composition
    Skills   []string
}

// Overrides the PrintDetails
func (d Developer) PrintDetails() {
    // Call Employee PrintDetails
    d.Employee.PrintDetails()
    fmt.Println("Technical Skills:")
    for _, v := range d.Skills {
        fmt.Println(v)
    }
}

type Manager struct {
    Employee  //type embedding for composition
    Projects  []string
    Locations []string
}

// Overrides the PrintDetails
func (m Manager) PrintDetails() {
    // Call Employee PrintDetails
    m.Employee.PrintDetails()
    fmt.Println("Projects:")
    for _, v := range m.Projects {
        fmt.Println(v)
    }
    fmt.Println("Managing teams for the locations:")
    for _, v := range m.Locations {
        fmt.Println(v)
    }
}

type Team struct {
    Name, Description string
    TeamMembers       []TeamMember
}

func (t Team) PrintTeamDetails() {
    fmt.Printf("Team: %s  - %s\n", t.Name, t.Description)
    fmt.Println("Details of the team members:")
    for _, v := range t.TeamMembers {
        v.PrintName()
        v.PrintDetails()
    }
}

func main() {
    steve := Developer{
        Employee{
            "Steve",
            "John",
            time.Date(1990, time.February, 17, 0, 0, 0, 0, time.UTC),
            "Software Engineer",
            "San Francisco",
        },
        []string{"Go", "Docker", "Kubernetes"},
    }
    irene := Developer{
        Employee{
            "Irene",
            "Rose",
            time.Date(1991, time.January, 13, 0, 0, 0, 0, time.UTC),
            "Software Engineer",
            "Santa Clara",
        },
        []string{"Go", "MongoDB"},
    }
    alex := Manager{
        Employee{
            "Alex",
            "Williams",
            time.Date(1979, time.February, 17, 0, 0, 0, 0, time.UTC),
            "Program Manger",
            "Santa Clara",
        },
        []string{"CRM", "e-Commerce"},
        []string{"San Francisco", "Santa Clara"},
    }

    // Create team
    team := Team{
        "Go",
        "Golang Engineering Team",
        []TeamMember{steve, irene, alex},
    }
    // Get details of Team
    team.PrintTeamDetails()
}

Listing 3-21.Example Program Demonstrates Interface with Type Embedding and Method Overriding

```

一个名为`Team`的结构被声明为代表一个雇员团队，团队成员的雇员由字段`TeamMembers`组织，字段的类型为`TeamMember`接口的切片。因为`TeamMembers`字段的类型使用了`TeamMember`接口的一部分，所以您可以提供`TeamMember`接口的任何实现作为值。类型`Employee`仅用于嵌入到`Developer`和`Manager`结构中，这些结构更多的是作为团队成员的雇员的具体实现。

```go
type Team struct {
    Name, Description string
    TeamMembers       []TeamMember
}

```

`Team`的`PrintTeamDetails`方法打印一个`Team`对象的信息。在`PrintTeamDetails`方法中，它遍历`TeamMembers`集合的元素，并调用`PrintName`和`PrintDetails`方法来获取每个团队成员的信息。

```go
func (t Team) PrintTeamDetails() {
    fmt.Printf("Team: %s  - %s\n", t.Name, t.Description)
    fmt.Println("Details of the team members:")
    for _, v := range t.TeamMembers {
        v.PrintName()
        v.PrintDetails()
    }
}

```

在`main`函数内部，通过提供实现了`TeamMember`接口的三个对象的值，创建了一个 team struct 实例。在三个`TeamMember`类型的对象中，两个是用`Developer`类型创建的，另一个是用`Manager`类型创建的。`TeamMembers`字段的值包含不同类型的值；所有对象的连接因素是`TeamMember`接口。您只需提供`TeamMember`接口的不同实现。最后调用`Team`结构的`PrintTeamDetails`方法来获取关于`Team`类型的值的信息。

```go
func main() {
    steve := Developer{
        Employee{
            "Steve",
            "John",
            time.Date(1990, time.February, 17, 0, 0, 0, 0, time.UTC),
            "Software Engineer",
            "San Francisco",
        },
        []string{"Go", "Docker", "Kubernetes"},
    }
    irene := Developer{
        Employee{
            "Irene",
            "Rose",
            time.Date(1991, time.January, 13, 0, 0, 0, 0, time.UTC),
            "Software Engineer",
            "Santa Clara",
        },
        []string{"Go", "MongoDB"},
    }
    alex := Manager{
        Employee{
            "Alex",
            "Williams",
            time.Date(1979, time.February, 17, 0, 0, 0, 0, time.UTC),
            "Program Manger",
            "Santa Clara",
        },
        []string{"CRM", "e-Commerce"},
        []string{"San Francisco", "Santa Clara"},
    }

    // Create team
    team := Team{
        "Go",
        "Golang Engineering Team",
        []TeamMember{steve, irene, alex},
    }
    // Get details of Team
    team.PrintTeamDetails()
}

```

运行该程序时，您应该会看到以下输出:

```go
Team: Go  - Golang Engineering Team
Details of the team members:

Steve John
Date of Birth: 1990-02-17 00:00:00 +0000 UTC, Job: Software Engineer, Location: San Francisco
Technical Skills:
Go
Docker
Kubernetes

Irene Rose
Date of Birth: 1991-01-13 00:00:00 +0000 UTC, Job: Software Engineer, Location: Santa Clara
Technical Skills:
Go
MongoDB

Alex Williams
Date of Birth: 1979-02-17 00:00:00 +0000 UTC, Job: Program Manger, Location: Santa Clara
Projects:
CRM
e-Commerce
Managing teams for the locations:
San Francisco
Santa Clara

```