# 五、使用标准库包

包是 Go 生态系统中非常重要的组成部分。Go 代码被组织成包，使你的 Go 程序具有可重用性和可组合性。Go 安装附带了许多可重用的包，称为标准库包。这些包扩展了 Go 语言，并为构建各种应用程序提供了可重用的库。它们可以帮助您快速构建应用程序，因为您不需要为许多常见功能编写自己的包。如果您想扩展标准库包，您可以创建自己的包，也可以获得 Go 开发者社区提供的第三方包。标准库包的功能非常丰富。您可以只使用标准库包来构建成熟的 web 应用程序，而不使用任何第三方包。本章介绍了如何使用标准库包来实现一些常见的功能，例如编码和解码 JavaScript 对象符号(JSON)对象、解析命令行标志、记录 Go 程序和归档文件。标准库包的文档可从 [`https://golang.org/pkg/`](https://golang.org/pkg/) 获得。

## 5-1.编码和解码 JSON

### 问题

您希望将 Go 类型的值编码成 JSON 对象，并将 JSON 对象解码成 Go 类型的值。

### 解决办法

标准库包`encoding/json`用于编码和解码 JSON 对象。

### 它是如何工作的

JSON 是一种数据交换格式，广泛用于 web 后端服务器与 web 和移动应用程序前端之间的通信。当您使用 Go 构建 RESTful 应用程序编程接口(API)时，您可能需要从 HTTP 请求体中解码 JSON 值，并将这些数据解析为 Go 值，并将 Go 值编码为 JSON 值以发送到 HTTP 响应。

#### 编码 JSON

`json`包的`Marshal`函数用于将 Go 值编码成 JSON 值。要使用`json`包，您必须将包`encoding/json`添加到导入列表中。

```go
import (
       "encoding/json"
)

```

下面是函数`Marshal`的签名:

```go
func Marshal(v interface{}) ([]byte, error)

```

函数`Marshal`返回两个值:作为`slice byte`的编码 JSON 数据和一个`error`值。

让我们声明一个 struct 类型来演示将 struct 类型的值解析到 JSON 中:

```go
type Employee struct {
        ID                            int
        FirstName, LastName, JobTitle string
}

```

下面的代码块创建了一个`Employee` struct 的实例，并将值解析成 JSON。

```go
emp := Employee{
                ID:        100,
                FirstName: "Shiju",
                LastName:  "Varghese",
                JobTitle:  "Architect",
        }
    // Encoding to JSON
    data, err := json.Marshal(emp)

```

函数`Marshal`返回`Employee`结构值的 JSON 编码。当您构建基于 JSON 的 RESTful APIs 时，您主要是将 struct 类型的值解析到 JSON 对象中。使用`Marshal`，您可以轻松地将 struct 类型的值编码为 JSON 值，这将帮助您快速构建基于 JSON 的 API。

#### 解码 JSON

`json`包的函数`Unmarshal`用于将 JSON 值解码成 Go 值。下面是函数`Unmarshal`的签名:

```go
func Unmarshal(data []byte, v interface{}) error

```

函数`Unmarshal`解析 JSON 编码的数据，并将结果存储到第二个参数中(`v interface{}`)。下面的代码块解码 JSON 数据，并将结果存储到`Employee`结构的值中:

```go
b := []byte(`{"ID":101,"FirstName":"Irene","LastName":"Rose","JobTitle":"Developer"}`)
var emp1 Employee
// Decoding JSON data into the value of Employee struct
err = json.Unmarshal(b, &emp1)

```

前面的语句解析变量`b`的 JSON 数据，并将结果存储到变量`emp1`中。JSON 数据是使用反引号作为原始字符串提供的。在反引号中，除了反引号之外，任何字符都是有效的。现在，您可以像读取普通结构值一样读取`Employee`结构的字段，如下所示:

```go
fmt.Printf("ID:%d, Name:%s %s, JobTitle:%s", emp1.ID, emp1.FirstName, emp1.LastName, emp1.JobTitle)

```

#### 示例:编码和解码

清单 [5-1](#Par22) 显示了一个示例程序，该程序演示了将 struct 类型的值编码到 JSON 对象中，以及将 JSON 对象解码成 struct 类型的值。

```go
package main

import (
        "encoding/json"
        "fmt"
)

// Employee struct
type Employee struct {
        ID                            int
        FirstName, LastName, JobTitle string
}

func main() {
        emp := Employee{
                ID:        100,
                FirstName: "Shiju",
                LastName:  "Varghese",
                JobTitle:  "Architect",
        }
    // Encoding to JSON
        data, err := json.Marshal(emp)
        if err != nil {
                fmt.Println(err.Error())
                return
        }
        jsonStr := string(data)
              fmt.Println("The JSON data is:")
        fmt.Println(jsonStr)

        b := []byte(`{"ID":101,"FirstName":"Irene","LastName":"Rose","JobTitle":"Developer"}`)
        var emp1 Employee
    // Decoding JSON data to a value of struct type
        err = json.Unmarshal(b, &emp1)
        if err != nil {
                fmt.Println(err.Error())
                return
        }
               fmt.Println("The Employee value is:")
        fmt.Printf("ID:%d, Name:%s %s, JobTitle:%s", emp1.ID, emp1.FirstName, emp1.LastName, emp1.JobTitle)
}

Listing 5-1.Encoding and Decoding of JSON with a Struct Type

```

运行该程序时，您应该会看到以下输出:

```go
The JSON data is:
{"ID":100,"FirstName":"Shiju","LastName":"Varghese","JobTitle":"Architect"}
The Employee value is:
ID:101, Name:Irene Rose, JobTitle:Developer

```

Note

当使用 struct 类型的值对 JSON 数据进行编码和解码时，必须将 struct 类型的所有字段指定为导出字段(标识符名称以大写字母开头),因为在调用`Marshal`和`Unmarshal`函数时,`json`包正在使用 struct 字段的值。

#### 结构标记

当您将 struct 类型的值编码到 JSON 中时，您可能需要在 JSON 编码中使用与 struct 类型的字段不同的字段。例如，您可以以大写字母开头来指定 struct 字段的名称，以将它们标记为导出字段，但是在 JSON 中，元素通常以小写字母开头。在这里，我们可以使用 struct 标记将 struct 字段的名称与 JSON 中的字段名称进行映射，以便在编码和解码 JSON 对象时使用。

下面是用 JSON 编码中要使用的标签和不同名称指定的`Employee`结构:

```go
type Employee struct {
        ID        int    `json:"id,omitempty"`

        FirstName string `json:"firstname"`

        LastName  string `json:"lastname"`

        JobTitle  string `json:"job"`

}

```

注意，反引号(`` ` ``)用于指定标签。在引号中，您将包`json`的元数据称为标签。在引号内，除了另一个反引号之外，任何字符都是有效的。结构字段`ID`用`id`标记，用于 JSON 表示。`omitempty`标志指定如果该字段有默认值，则该字段不包含在 JSON 表示中。如果您没有为`Employee`结构的`ID`字段提供值，那么当您将`Employee`值解析到 JSON 时，JSON 对象的输出不包括`id`字段。对于 JSON 数据，`Employee`结构的所有字段都用不同的名称标记。

如果你想从结构中跳过字段，你可以给标签名为`"-"`。这里显示的`User`结构指定在编码和解码 JSON 对象时必须跳过字段`Password`:

```go
type User struct {
    UserName string `json:"user"`
    Password string `json:"-"`

}

```

#### 示例:使用 Struct 标记进行编码和解码

清单 [5-2](#Par33) 展示了一个示例程序，演示了用 struct 标签对 JSON 对象进行编码和解码。

```go
package main

import (
        "encoding/json"
        "fmt"
)

// Employee struct with struct tags
type Employee struct {
        ID        int    `json:"id,omitempty"`
        FirstName string `json:"firstname"`
        LastName  string `json:"lastname"`
        JobTitle  string `json:"job"`
}

func main() {
        emp := Employee{
                FirstName: "Shiju",
                LastName:  "Varghese",
                JobTitle:  "Architect",
        }
        // Encoding to JSON
        data, err := json.Marshal(emp)
        if err != nil {
                fmt.Println(err.Error())
                return
        }
        jsonStr := string(data)
        fmt.Println("The JSON data is:")
        fmt.Println(jsonStr)

        b := []byte(`{"id":101,"firstname":"Irene","lastname":"Rose","job":"Developer"}`)
        var emp1 Employee
        // Decoding JSON to a struct type
        err = json.Unmarshal(b, &emp1)
        if err != nil {
                fmt.Println(err.Error())
                return
        }
        fmt.Println("The Employee value is:")
        fmt.Printf("ID:%d, Name:%s %s, JobTitle:%s", emp1.ID, emp1.FirstName, emp1.LastName, emp1.JobTitle)
}

Listing 5-2.Encoding and Decoding of JSON with Struct Tags

```

运行该程序时，您应该会看到以下输出:

```go
The JSON data is:
{"firstname":"Shiju","lastname":"Varghese","job":"Architect"}
The Employee value is:
ID:101, Name:Irene Rose, JobTitle:Developer

```

`Employee`结构用字段名称标记，用于 JSON 对象。在没有指定`ID`字段的情况下创建了`Employee`结构的值，因此在将`Employee`结构的值编码到 JSON 中时，JSON 对象不包括`id`字段。JSON 输出还显示了在 struct 声明中标记的相应 JSON 字段名。当我们解码 JSON 对象时，`id`字段不为空，因此它被解析到`Employee`结构的`ID`字段中。

## 5-2.使用命令行标志

### 问题

您希望解析命令行标志，以便为 Go 程序提供一些值。

### 解决办法

标准库包`flag`用于解析命令行标志。

### 它是如何工作的

有时，在运行程序时，您可能需要通过命令行从最终用户那里接收值。这是构建命令行应用程序时的一个基本特性。命令行选项，也称为标志，可用于在运行程序时向程序提供值。标准库包`flag`提供了解析命令行标志的函数。包`flag`提供了使用`flag.String()`、`flag.Bool()`和`flag.Int()`解析`string`、`integer`和`boolean`值的函数。

要使用`flag`包，您必须将其添加到导入列表中:

```go
import (
       "flag"
)

```

清单 [5-3](#Par43) 显示了一个示例程序，演示了如何在 Go 程序中定义标志。

```go
package main

import (
        "flag"
        "fmt"
)

func main() {

        fileName := flag.String("filename", "logfile", "File name for the log file")
        logLevel := flag.Int("loglevel", 0, "An integer value for Level (0-4)")
        isEnable := flag.Bool("enable", false, "A boolean value for enabling log options")
        var num int
        // Bind the flag to a variable.
        flag.IntVar(&num, "num", 25, "An integer value")

        // Parse parses flag definitions from the argument list.
        flag.Parse()
        // Get the values from pointers
        fmt.Println("filename:", *fileName)
        fmt.Println("loglevel:", *logLevel)
        fmt.Println("enable:", *isEnable)
        // Get the value from a variable
        fmt.Println("num:", num)
        // Args returns the non-flag command-line arguments.
        args := flag.Args()
        if len(args) > 0 {
                fmt.Println("The non-flag command-line arguments are:")
                // Print the arguments
                for _, v := range args {
                        fmt.Println(v)
                }
        }

}

Listing 5-3.Defining Flags Using Package flag

```

函数`flag.String`用于定义通过命令行获取`string`值的标志。

```go
fileName := flag.String("filename", "logfile", "File name for the log file")

```

前面的语句声明了一个`string`标志，标志名为`filename`，并提供了一个默认值`"logfile "`。`filename`标志`(-filename`的用户输入存储在指针`fileName`中，类型为`*string`。第三个参数描述了标志的用法。函数`flag.Bool()`和`flag.Int()`用于声明`boolean`和`integer`值的标志。

```go
logLevel := flag.Int("loglevel", 0, "An integer value for Level (0-4)")
isEnable := flag.Bool("enable", false, "A boolean value for enabling log options")

```

如果您想将标志绑定到一个现有的变量，您可以使用函数`flag.IntVar`、`flag.BoolVar`和`flag.StringVar`。下面的代码块将标志`num` ( `-num`)绑定到`integer`变量`num`。

```go
var num int
// Bind the flag to a variable.
flag.IntVar(&num, "num", 25, "An integer value")

```

函数`flag.Parse()`从命令行解析标志定义。因为函数`flag.String()`、`flag.Bool()`和`flag.Int()`是返回指针，所以我们解引用这些指针来获取值。

```go
fmt.Println("name:", *fileName)
fmt.Println("num:", *logLevel)
fmt.Println("enable:", *isEnable)

```

函数`flag.IntVar`返回一个`integer`值，而不是一个指针，这样就可以在不引用指针的情况下读取该值。

```go
fmt.Println("num:", num)

```

包`flag`提供了一个名为`Args`的函数，可以用来读取非 flag 命令行参数。如果您提供非 flag 命令行参数，这个函数调用将返回一个`string`的`slice`。命令行参数位于命令行标志之后。如果用户提供的话，命令行参数会打印到控制台中。

```go
args := flag.Args()
        if len(args) > 0 {
                fmt.Println("The non-flag command-line arguments are:")
                // Print the arguments
                for _, v := range args {
                        fmt.Println(v)
                }
        }

```

让我们构建程序，并使用不同的命令行选项运行它:

```go
$ go build

```

首先，让我们通过提供所有的标志和参数来运行程序。

```go
$ ./ cmdflags -filename=applog -loglevel=2 -enable -num=50 10 20 30 test
filename: applog
loglevel: 2
enable: true
num: 50
The non-flag command-line arguments are:
10
20
30
test

```

必须在给出标志后提供非标志命令行参数。标志`-h`或`--help`为命令行程序的使用提供帮助。该帮助文本将由程序中定义的标志定义生成。让我们通过提供`-h`标志来运行程序。

```go
$ ./ cmdflags -h
Usage of cmdflags:
  -enable
        A boolean value for enabling log options
  -filename string
        File name for the log file (default "logfile")
  -loglevel int
        An integer value for Level (0-4)
  -num int
        An integer value (default 25)

```

现在让我们通过提供几个不带非标志参数的标志来运行程序:

```go
$ ./ cmdflags -filename=applog -loglevel=1
filename: applog
loglevel: 1
enable: false
num: 25

```

如果用户没有为标志提供值，将采用默认值。

## 5-3.记录 Go 程序

### 问题

您希望为您的 Go 程序实现日志记录。

### 解决办法

标准库包`log`提供了一个基本的日志基础设施，可以用来记录你的 Go 程序。

### 它是如何工作的

尽管有许多第三方包可用于日志记录，但如果您想继续使用标准库或使用简单的包，标准库包`log`应该是您的选择。包`log`允许你将日志信息写入所有支持`io.Writer`接口的标准输出设备。struct type `log.Logger`是包`log,`中的主要组件，它提供了几种日志记录方法，也支持格式化日志数据。

要使用包`log`，您必须将其添加到导入列表:

```go
import (
       "log"
)

```

#### 示例:一个基本的记录器

清单 [5-4](#Par71) 显示了一个使用`log.Logger`类型提供基本日志实现的示例程序。日志消息分为跟踪、信息、警告和错误，每个日志类别使用四个`log.Logger`对象。

```go
package main

import (
        "errors"
        "io"
        "io/ioutil"
        "log"
        "os"
)

// Package level variables, which are pointers to log.Logger.
 var (
        Trace   *log.Logger
        Info    *log.Logger
        Warning *log.Logger
        Error   *log.Logger
)

// initLog initializes log.Logger objects
func initLog(
        traceHandle io.Writer,
        infoHandle io.Writer,
        warningHandle io.Writer,
        errorHandle io.Writer) {

        // Flags for defineing the logging properties, to log.New
        flag := log.Ldate | log.Ltime | log.Lshortfile

        // Create log.Logger objects
        Trace = log.New(traceHandle, "TRACE: ", flag)
        Info = log.New(infoHandle, "INFO: ", flag)
        Warning = log.New(warningHandle, "WARNING: ", flag)
        Error = log.New(errorHandle, "ERROR: ", flag)

}

func main() {
        initLog(ioutil.Discard, os.Stdout, os.Stdout, os.Stderr)
        Trace.Println("Main started")
        loop()
        err := errors.New("Sample Error")
        Error.Println(err.Error())
        Trace.Println("Main completed")
}
func loop() {
        Trace.Println("Loop started")
        for i := 0; i < 10; i++ {
                Info.Println("Counter value is:", i)
        }
        Warning.Println("The counter variable is not being used")
        Trace.Println("Loop completed")
}

Listing 5-4.A Basic Logging Implementation with Categorized Logging for Trace, Information, Warning, and Error Messages

```

为跟踪、信息、警告和错误的分类日志记录声明了四个指向类型`log.Logger`的指针。通过调用函数`initLog`来创建`log.Logger`对象，该函数接收接口`io.Writer`的参数来设置日志消息的目的地。

```go
// Package level variables, which are pointers to log.Logger.
var (
        Trace   *log.Logger
        Info    *log.Logger
        Warning *log.Logger
        Error   *log.Logger
)

// initLog initializes log.Logger objects
func initLog(
        traceHandle io.Writer,
        infoHandle io.Writer,
        warningHandle io.Writer,
        errorHandle io.Writer) {

        // Flags for defining the logging properties, to log.New
        flag := log.Ldate | log.Ltime | log.Lshortfile

        // Create log.Logger objects
        Trace = log.New(traceHandle, "TRACE: ", flag)
        Info = log.New(infoHandle, "INFO: ", flag)
        Warning = log.New(warningHandle, "WARNING: ", flag)
        Error = log.New(errorHandle, "ERROR: ", flag)

}

```

函数`log.New`创建一个新的`log.Logger`。在函数`New`中，第一个参数设置日志数据的目的地，第二个参数设置出现在每个生成的日志行开头的前缀，第三个参数定义日志属性。给定的日志记录属性在日志数据中提供日期、时间和短文件名。日志数据可以写入任何支持接口`io.Writer`的目的地。从功能`main`调用功能`initLog`。

```go
initLog(ioutil.Discard, os.Stdout, os.Stdout, os.Stderr)

```

将`ioutil.Discard`提供给 Trace 的目的地，这是一个空设备，因此这个目的地的所有日志写调用都将成功，而无需做任何事情。`os.Stdout`是给目的地的信息和警告，因此该目的地的所有日志写调用都将出现在控制台窗口中。将`os.Stderr`赋予错误的目的地，以便该目的地的所有日志写调用将作为标准错误出现在控制台窗口中。在这个示例程序中，`Logger`跟踪、信息、警告和错误的对象用于记录消息。因为跟踪的目的地被配置为`ioutil.Discard`，日志数据不会出现在控制台窗口中。

您应该会看到类似如下的输出:

```go
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 0
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 1
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 2
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 3
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 4
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 5
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 6
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 7
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 8
INFO: 2016/06/11 18:47:28 main.go:48: Counter value is: 9
WARNING: 2016/06/11 18:47:28 main.go:50: The counter variable is not being used
ERROR: 2016/06/11 18:47:28 main.go:42: Sample Error

```

#### 示例:可配置的记录器

在前面的例子中，日志数据被写入`Stdout`和`Stderr`接口。然而，当您开发真实世界的应用程序时，您可能会使用持久化存储作为日志数据的目的地。您可能还需要一个可配置的选项来指定跟踪、信息、警告或错误的日志级别。这使您可以随时更改日志级别。例如，您可能将日志级别设置为跟踪，但在将应用程序投入生产时，您可能不需要跟踪级别的日志。

清单 [5-5](#Par81) 显示了一个示例程序，它提供了一个日志基础设施，允许您将日志级别配置为跟踪、信息、警告或错误，然后将日志数据写入一个文本文件。可以使用命令行标志来配置日志级别选项。

```go
package main

import (
        "io"
        "io/ioutil"
        "log"
        "os"
)

const (
        // UNSPECIFIED logs nothing
        UNSPECIFIED Level = iota // 0 :
        // TRACE logs everything
        TRACE // 1
        // INFO logs Info, Warnings and Errors
        INFO // 2
        // WARNING logs Warning and Errors
        WARNING // 3
        // ERROR just logs Errors
        ERROR // 4
)

// Level holds the log level.
type Level int

// Package level variables, which are pointers to log.Logger.
var (
        Trace   *log.Logger
        Info    *log.Logger
        Warning *log.Logger
        Error   *log.Logger
)

// initLog initializes log.Logger objects
func initLog(
        traceHandle io.Writer,
        infoHandle io.Writer,
        warningHandle io.Writer,
        errorHandle io.Writer,
        isFlag bool) {

        // Flags for defining the logging properties, to log.New
        flag := 0
        if isFlag {
                flag = log.Ldate | log.Ltime | log.Lshortfile
        }

        // Create log.Logger objects.
        Trace = log.New(traceHandle, "TRACE: ", flag)
        Info = log.New(infoHandle, "INFO: ", flag)
        Warning = log.New(warningHandle, "WARNING: ", flag)
        Error = log.New(errorHandle, "ERROR: ", flag)

}

// SetLogLevel sets the logging level preference
func SetLogLevel(level Level) {

        // Creates os.*File, which has implemented io.Writer interface
        f, err := os.OpenFile("logs.txt", os.O_RDWR|os.O_CREATE|os.O_APPEND, 0666)
        if err != nil {
                log.Fatalf("Error opening log file: %s", err.Error())
        }

        // Calls function initLog by specifying log level preference.
        switch level {
        case TRACE:
                initLog(f, f, f, f, true)
                return

        case INFO:
                initLog(ioutil.Discard, f, f, f, true)
                return

        case WARNING:
                initLog(ioutil.Discard, ioutil.Discard, f, f, true)
                return
        case ERROR:
                initLog(ioutil.Discard, ioutil.Discard, ioutil.Discard, f, true)
                return

        default:
                initLog(ioutil.Discard, ioutil.Discard, ioutil.Discard, ioutil.Discard, false)
                f.Close()
                return

        }
}

Listing 5-5.A Logging Infrastructure with an Option to Set the Log Level and Write Log Data into a Text File, in logger.go

```

`logger.go`源提供两个功能:`initLog`和`SetLogLevel`。函数`SetLogLevel`通过调用标准库包`os`的函数`OpenFile`来创建文件对象，然后调用函数`initLog`通过提供日志级别首选项来初始化`Logger`对象。它打开带有指定标志的命名文件。函数`initLog`根据函数提供的日志首选项创建`Logger`对象。

声明常量变量是为了指定不同级别的日志级别首选项。标识符`iota`用于构造一组相关的常数；在这里，它用于组织应用程序中可用的日志级别，这将产生一个自动递增的`integer`常量。每当`const`出现在源代码中时，它将值重置为 0，并在常量声明中的每个值之后递增。

```go
const (
        // UNSPECIFIED logs nothing
        UNSPECIFIED Level = iota // 0 :
        // TRACE logs everything
        TRACE // 1
        // INFO logs Info, Warnings and Errors
        INFO // 2
        // WARNING logs Warning and Errors
        WARNING // 3
        // ERROR just logs Errors
        ERROR // 4
)

// Level holds the log level.
type Level int

```

在许多编程语言中，枚举或简单的枚举是声明具有相似行为的常数的惯用方式。与某些编程语言不同，Go 不支持使用关键字来声明枚举。在 Go 中声明枚举的惯用方式是用`iota`声明常量。这里，名为`Level`的类型和类型`int`用于指定常量的类型。常量`UNSPECIFIED`的值重置为 0，然后它自动递增每个常量声明，1 代表`TRACE`，2 代表`INFO`，依此类推。

清单 [5-6](#Par87) 显示了一个使用在`logger.go`中实现的日志基础设施的 Go 源文件(参见清单 [5-5](#Par81) )。

```go
package main

import (
        "errors"
        "flag"
)

func main() {
        // Parse log level from command line
        logLevel := flag.Int("loglevel", 0, "an integer value (0-4)")
        flag.Parse()
        // Calling the SetLogLevel with the command-line argument
        SetLogLevel(Level(*logLevel))
        Trace.Println("Main started")
        loop()
        err := errors.New("Sample Error")
        Error.Println(err.Error())
        Trace.Println("Main completed")
}
// A simple function for the logging demo
func loop() {
        Trace.Println("Loop started")
        for i := 0; i < 10; i++ {
                Info.Println("Counter value is:", i)
        }
        Warning.Println("The counter variable is not being used")
        Trace.Println("Loop completed")
}

Listing 5-6.Logging Demo in main.go, Using logger.go

```

在函数`main`中，从命令行标志接受日志级别首选项的值，并调用`logger.go`的函数`SetLogLevel`通过指定日志级别首选项来创建`Logger`对象。

```go
logLevel := flag.Int("loglevel", 0, "an integer value (0-4)")
flag.Parse()
// Calling the SetLogLevel with the command-line argument
SetLogLevel(Level(*logLevel))

```

在本例中，使用`Logger`对象记录跟踪、信息、警告和错误。让我们通过为 Trace 提供日志级别首选项(值 1)来运行程序。

```go
$ go build
$ ./log -loglevel=1

```

这会将日志数据写入名为`logs.txt`的文本文件。要跟踪的日志级别写入了`Trace`、`Information`、`Warning`和`Error`的日志数据。您应该会在`logs.txt`中看到类似如下的日志数据。

```go
TRACE: 2016/06/13 22:04:28 main.go:14: Main started
TRACE: 2016/06/13 22:04:28 main.go:23: Loop started
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 0
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 1
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 2
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 3
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 4
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 5
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 6
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 7
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 8
INFO: 2016/06/13 22:04:28 main.go:25: Counter value is: 9
WARNING: 2016/06/13 22:04:28 main.go:27: The counter variable is not being used
TRACE: 2016/06/13 22:04:28 main.go:28: Loop completed
ERROR: 2016/06/13 22:04:28 main.go:17: Sample Error
TRACE: 2016/06/13 22:04:28 main.go:18: Main completed

```

让我们通过指定信息的日志级别来运行程序(`loglevel`的值为 2)。

```go
$ ./log -loglevel=2

```

您应该会看到类似下面的日志数据附加到`logs.txt`中。

```go
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 0
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 1
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 2
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 3
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 4
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 5
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 6
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 7
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 8
INFO: 2016/06/13 22:13:25 main.go:25: Counter value is: 9
WARNING: 2016/06/13 22:13:25 main.go:27: The counter variable is not being used
ERROR: 2016/06/13 22:13:25 main.go:17: Sample Error

```

因为我们将日志级别指定为 Information，所以 Information、Warning 和 Error 的日志数据被附加到输出文件`logs.txt`中，但是 Trace 的日志数据被写入空设备中。

## 5-4.以 Tar 和 Zip 格式存档文件

### 问题

你想读写 tar 和 zip 格式的文件。

### 解决办法

标准库包`archive`包含两个子包——包`archive/tar`和包`archive/zip,`，用于读写 tar 和 zip 格式的归档文件。

### 它是如何工作的

标准库包`archive`支持以两种文件格式归档文件。为了支持 tar 和 zip 格式的归档功能，它提供了两个独立的包:`archive/tar`和`archive/zip`。`archive/tar`和`archive/zip`包分别为 tar 和 zip 格式的读写提供支持。

#### io。作家和木卫一。阅读器界面

在开始写入和读取归档文件之前，让我们先来看看`io.Writer`和`io.Reader`接口。标准库包`io`提供了执行 I/O 操作的基本接口。包`io`的`Writer`接口为写操作提供了一个抽象。`Writer`接口声明了一个名为`Write`的方法，该方法接受一个值`byte slice`作为参数。

下面是接口`io.Writer`的声明:

```go
type Writer interface {
        Write(p []byte) (n int, err error)
}

```

下面是`Write`方法的 Go 文档:

`Write`将`len(p)`字节从`p`写入底层数据流。它返回从`p (0 <= n <= len(p))`写入的字节数，以及遇到的导致写入提前停止的任何`error`。`Write`如果返回`n < len(p)`则必须返回一个`non-nil error`。写入不得修改`slice`数据，即使是暂时的。

包`io`的`Reader`接口为读操作提供了一个抽象。`Reader`接口声明了一个名为`Read`的方法，该方法接受一个值`byte slice`作为参数。

下面是`io.Reader`接口的声明:

```go
type Reader interface {
        Read(p []byte) (n int, err error)
}

```

下面是关于`Read`方法的 Go 文档:

Read 将最多`len(p)`个字节读入`p`。它返回读取的字节数(`0 <= n <= len(p)`和遇到的任何`error`)。即使`Read`返回`n < len(p`，它也可能在调用过程中使用所有的`p`作为暂存空间。如果有些数据可用，但没有`len(p)`字节，`Read`通常会返回可用的数据，而不是等待更多数据。当`Read`在成功读取`n > 0`字节后遇到`error`或文件结束条件时，它返回读取的字节数。它可以从同一个调用返回(`non-nil` ) `error`，或者从后续调用返回`error`(和`n == 0`)。这种一般情况的一个例子是，在输入流末尾返回非零字节数的`Reader`可能返回`err == EOF`或`err == nil`。下一个`Read`应该会返回`0, EOF`。

当您读写归档文件时，您将利用`io.Writer`和`io.Reader`接口。

#### 写入和读取 Tar 文件

包`archive/tar`用于读写 tar 文件。Tar(磁带归档)文件是在基于 Unix 的系统中使用的归档文件。tar 存档的文件后缀是`.tar`。`tar` Unix shell 命令从多个指定的文件中创建一个归档文件，或者从归档文件中提取文件。要使用包`archive/tar`，您必须将它添加到导入列表中:

```go
import (
       "archive/tar"
)

```

结构类型`tar.Writer`用于将文件写入 tar 文件。通过调用接受类型为`io.Writer`的值的函数`tar.NewWriter`来创建`Writer`对象，您可以将 tar 存档文件作为类型为`os.File`的对象传递给该函数，以写入所提供的 tar 文件。结构类型`os.File`已经实现了`io.Writer`接口，因此它可以用作调用函数`tar.NewWriter`的参数。

结构类型`tar.Reader`用于从 tar 文件中读取文件。通过调用函数`tar.NewReader`来创建`Reader`对象，该函数接受类型为`io.Reader`的值作为参数，您可以将 tar 存档文件作为类型为`os.File`的对象传递给该参数，以读取 tar 文件的内容。结构类型`os.File`已经实现了接口`io.Reader`，因此它可以用作调用函数`tar.NewReader`的参数。

清单 [5-7](#Par118) 显示了一个示例程序，它演示了如何通过将两个文件写入一个 tar 文件，然后通过遍历 tar 文件并读取每个文件的内容来读取 tar 文件，从而对文件进行归档。

```go
package main

import (
        "archive/tar"
        "fmt"
        "io"
        "log"
        "os"
)

// addToArchive writes a given file into a .tar file
// Returns nill if the operation is succeeded
func addToArchive(filename string, tw *tar.Writer) error {
        // Open the file to archive into tar file.
        file, err := os.Open(filename)
        if err != nil {
                return err
        }
        defer file.Close()
        // Get the FileInfo struct that describes the file.
        fileinfo, err := file.Stat()
        // Create a pointer to tar.Header struct
        hdr := &tar.Header{
                ModTime: fileinfo.ModTime(),            // modified time
                Name:    filename,                      // name of header
                Size:    fileinfo.Size(),               // length in bytes
                Mode:    int64(fileinfo.Mode().Perm()), // permission and mode bits
        }
        // WriteHeader writes tar.Header and prepares to accept the file's contents.
        if err := tw.WriteHeader(hdr); err != nil {
                return err
        }
        // Write the file contents to the tar file.
        copied, err := io.Copy(tw, file)
        if err != nil {
                return err
        }
        // Check the size of copied file with the source file.
        if copied < fileinfo.Size() {
                return fmt.Errorf("Size of the copied file doesn't match with source file %s: %s", filename, err)
        }
        return nil
}

// archiveFiles archives a group of given files into a tar file.
func archiveFiles(files []string, archive string) error {
        // Flags for open the tar file.
        flags := os.O_WRONLY | os.O_CREATE | os.O_TRUNC
        // Open the tar file
        file, err := os.OpenFile(archive, flags, 0644)
        if err != nil {
                return err
        }
        defer file.Close()
        // Creates a new Writer writing to given file object.
        // Writer provides sequential writing of a tar archive in POSIX.1 format.
        tw := tar.NewWriter(file)
        defer tw.Close()
        // Iterate through the files to write each file into the tar file.
        for _, filename := range files {
                // Write the file into tar file.
                if err := addToArchive(filename, tw); err != nil {
                        return err
                }
        }
        return nil
}

// readArchive reads the file contents from tar file.
func readArchive(archive string) error {
        // Open the tar archive file.
        file, err := os.Open(archive)
        if err != nil {
                return err
        }
        defer file.Close()
        // Create the tar.Reader to read the tar archive.
        // A Reader provides sequential access to the contents of a tar archive.
        tr := tar.NewReader(file)
        // Iterate through the files in the tar archive.
        for {
                hdr, err := tr.Next()
                if err == io.EOF {
                        // End of tar archive
                        break
                }
                if err != nil {
                        return err
                }
                size := hdr.Size
                contents := make([]byte, size)
                read, err := io.ReadFull(tr, contents)
                // Check the size of file contents
                if int64(read) != size {
                        return fmt.Errorf("Size of the opened file doesn't match with the file %s", hdr.Name)
                }
                fmt.Printf("Contents of the file %s:\n", hdr.Name)
                // Writing the file contents into Stdout.
                fmt.Fprintf(os.Stdout, "\n%s", contents)
        }
        return nil
}

func main() {
        // Name of the tar file
        archive := "source.tar"
        // Files to be archived in tar format
        files := []string{"main.go", "readme.txt"}
        // Archive files into tar format
        err := archiveFiles(files, archive)
        if err != nil {
                log.Fatalf("Error while writing to tar file:%s", err)
        }
        // Archiving is successful.
        fmt.Println("The tar file source.tar has been created")
        // Read the file contents of tar file
        err = readArchive(archive)
        if err != nil {
                log.Fatalf("Error while reading the tar file:%s", err)
        }
}

Listing 5-7.Writing and Reading a Tar File

```

在函数`main`中，声明了一个变量`archive`来为 tar 文件提供文件名。声明一个变量`files`来提供文件名作为`string slice`来将提供的文件写入 tar 文件。调用函数`archiveFiles`来归档文件，调用另一个函数`readArchive`来读取 tar 文件的内容，该文件是使用函数`archiveFiles`写入的。

```go
func main() {
        // Name of the tar file
        archive := "source.tar"
        // Files to be archived in tar format
        files := []string{"main.go", "readme.txt"}
        // Archive files into tar format
        err := archiveFiles(files, archive)
        if err != nil {
                log.Fatalf("Error while writing to tar file:%s", err)
        }
        // Archiving is successful.
        fmt.Println("The tar file source.tar has been created")
        // Read the file contents of tar file
        err = readArchive(archive)
        if err != nil {
                log.Fatalf("Error while reading the tar file:%s", err)
        }
}

```

在函数`archiveFiles`内部，通过打开 tar 文件创建一个`os.File`对象，然后通过向函数`tar.NewWriter`传递一个`File`对象来创建一个新的`tar.Writer`。`Writer`用于将文件写入 tar 文件。

```go
// Open the tar file
file, err := os.OpenFile(archive, flags, 0644)
if err != nil {
        return err
}
defer file.Close()
// Create a new Writer writing to given file object.
// Writer provides sequential writing of a tar archive in POSIX.1 format.
tw := tar.NewWriter(file)

```

要将文件集合写入 tar 文件，您需要遍历变量`files`，它将文件名保存为值`string slice`，并调用函数`addToArchive`将提供的文件写入 tar 文件。

```go
for _, filename := range files {
        // Write the file into tar file.
        if err := addToArchive(filename, tw); err != nil {
                return err
        }
}

```

函数`addToArchive`使用`tar.Writer`将提供的文件写入 tar 文件。为了向 tar 文件写入一个新文件，通过提供`tar.Header`的值来调用`tar.Writer`对象的函数`WriteHeader`。然后它调用`io.Copy`将文件的数据写入 tar 文件。值`tar.Header`包含正在写入 tar 文件的文件的元数据。

```go
file, err := os.Open(filename)
if err != nil {
        return err
}
defer file.Close()
// Get the FileInfo struct that describes the file.
fileinfo, err := file.Stat()
// Create a pointer to tar.Header struct
hdr := &tar.Header{
        ModTime: fileinfo.ModTime(),            // modified time
        Name:    filename,                      // name of header
        Size:    fileinfo.Size(),               // length in bytes
        Mode:    int64(fileinfo.Mode().Perm()), // permission and mode bits
}
// WriteHeader writes tar.Header and prepares to accept the file's contents.
if err := tw.WriteHeader(hdr); err != nil {
        return err
}
// Write the file contents to the tar file.
copied, err := io.Copy(tw, file)

```

函数`readArchive`用于读取 tar 文件的文件内容。指向`tar.Reader`的指针用于读取 tar 文件，它是通过调用函数`tar.NewReader`并传递值`os.File`来创建的。

```go
// Open the tar archive file.
file, err := os.Open(archive)
if err != nil {
        return err
}
defer file.Close()
// Create the tar.Reader to read the tar archive.
// A Reader provides sequential access to the contents of a tar archive.
tr := tar.NewReader(file)

```

使用`tar.Reader`遍历 tar 文件中的文件，并读取写入`os.Stdout`的内容。`tar.Reader`的函数`Next`前进到文件中的下一个条目，并在文件末尾返回一个`io.EOF`的`error`值。当对函数`Next`的调用返回`io.EOF`时，您可以退出读取操作，因为这表明您已经遍历了所有文件内容并到达了文件的末尾。

```go
// Iterate through the files in the tar archive.
for {
        hdr, err := tr.Next()
        if err == io.EOF {
                // End of tar archive
                fmt.Println("end")
                break
        }
        if err != nil {
                return err
        }
        size := hdr.Size
        contents := make([]byte, size)
        read, err := io.ReadFull(tr, contents)
        // Check the size of file contents
        if int64(read) != size {
                return fmt.Errorf("Size of the opened file doesn't match with the file %s", hdr.Name)
        }
        // hdr.Name returns the file name.
        fmt.Printf("Contents of the file %s:\n", hdr.Name)
        // Writing the file contents into Stdout.
        fmt.Fprintf(os.Stdout, "\n%s", contents)
}

```

在这个例子中，您试图将源文件`main.go`和`readme.txt`归档到`source.tar`文件中。当您运行程序时，您应该看到应用程序目录中的归档文件`source.tar`作为写操作的输出，文件`main.go`和`readme.txt`的内容作为读操作的输出。

#### 编写和读取 Zip 文件

`archive/zip`包用于读写 zip 文件。要使用包`archive/zip`，您必须将其添加到导入列表中:

```go
import (
       "archive/zip"
)

```

包`archive/zip`提供了与`package archive/tar`相似的功能，用包`zip`读写 zip 文件的过程与处理 tar 文件的过程相似。struct type `zip.Writer`用于将文件写入 zip 文件。通过调用接受类型为`io.Writer`的值的函数`zip.NewWriter`来创建一个新的`zip.Writer`。

结构类型`zip.ReadCloser`可用于从 zip 文件中读取文件。通过调用函数`zip.OpenReader`可以创建`Reader`对象，该函数将打开 name 给出的 zip 文件并返回一个`zip.ReadCloser`。包`zip`还提供了一个类型`Reader`；通过调用函数`zip.NewReader`创建一个新的`Reader`。

清单 [5-8](#Par137) 显示了一个示例程序，它演示了如何通过将两个文件写入一个 zip 文件，然后通过遍历 zip 文件中包含的文件并读取每个文件的内容来读取 zip 文件，从而对文件进行归档。

```go
package main

import (
        "archive/zip"
        "fmt"
        "io"
        "log"
        "os"
)

// addToArchive writes a given file into a zip file.
func addToArchive(filename string, zw *zip.Writer) error {
        // Open the given file to archive into a zip file.
        file, err := os.Open(filename)
        if err != nil {
                return err
        }
        defer file.Close()
        // Create adds a file to the zip file using the given name/
        // Create returns a io.Writer to which the file contents should be written.
        wr, err := zw.Create(filename)
        if err != nil {
                return err
        }
        // Write the file contents to the zip file.
        if _, err := io.Copy(wr, file); err != nil {
                return err
        }
        return nil
}

// archiveFiles archives a group of given files into a zip file.
func archiveFiles(files []string, archive string) error {
        flags := os.O_WRONLY | os.O_CREATE | os.O_TRUNC
        // Open the tar file
        file, err := os.OpenFile(archive, flags, 0644)
        if err != nil {
                return err
        }
        defer file.Close()
        // Create zip.Writer that implements a zip file writer.
        zw := zip.NewWriter(file)
        defer zw.Close()
        // Iterate through the files to write each file into the zip file.
        for _, filename := range files {
                // Write the file into tar file.
                if err := addToArchive(filename, zw); err != nil {
                        return err
                }
        }
        return nil
}

// readArchive reads the file contents from tar file.
func readArchive(archive string) error {
        // Open the zip file specified by name and return a ReadCloser.
        rc, err := zip.OpenReader(archive)
        if err != nil {
                return err
        }
        defer rc.Close()
        // Iterate through the files in the zip file to read the file contents.
        for _, file := range rc.File {
                frc, err := file.Open()
                if err != nil {
                        return err
                }
                defer frc.Close()
                fmt.Fprintf(os.Stdout, "Contents of the file %s:\n", file.Name)
                // Write the contents into Stdout
                copied, err := io.Copy(os.Stdout, frc)
                if err != nil {
                        return err
                }
                // Check the size of the file.
                if uint64(copied) != file.UncompressedSize64 {
                        return fmt.Errorf("Length of the file contents doesn't match with the file %s", file.Name)
                }
                fmt.Println()
        }
        return nil
}

func main() {
        // Name of the zip file
        archive := "source.zip"
        // Files to be archived in zip format.
        files := []string{"main.go", "readme.txt"}
        // Archive files into zip format.
        err := archiveFiles(files, archive)
        if err != nil {
                log.Fatalf("Error while writing to zip file:%s\n", err)
        }
        // Read the file contents of tar file.
        err = readArchive(archive)
        if err != nil {
                log.Fatalf("Error while reading the zip file:%s\n", err)

        }
}

Listing 5-8.Writing and Reading a Zip File

```

这个例子类似于清单 8-7，只是在读写 tar 和 zip 格式文件的实现上有一些不同。当您运行程序时，您应该看到应用程序目录中的归档文件`source.zip`作为写操作的输出，文件`main.go`和`readme.txt`的内容作为读操作的输出。