# 九、模板

大多数服务器端语言都有一种机制，主要是获取静态页面并插入动态生成的组件，比如一个项目列表。典型的例子是 Java 服务器页面中的脚本、PHP 脚本和许多其他脚本。Go 在`template`包中采用了相对简单的脚本语言。

该包设计为将文本作为输入，并基于使用对象的值转换原始文本来输出不同的文本。不像 JSP 或类似的，它并不局限于 HTML 文件，但它很可能在那里找到最大的用途。我们首先描述`text/template`包，然后描述`html/template`包。

原始源称为模板，将由不变地传输的文本和可以作用于并改变文本的嵌入命令组成。命令由`{{ ... }}`分隔，类似于 JSP 命令`<%= ... =%>`和 PHP 的`<?php ... ?>`。

## 插入对象值

模板应用于 Go 对象。来自 Go 对象的字段可以插入到模板中，您可以“挖掘”对象以找到子字段，等等。当前对象表示为光标`.`，因此要将当前对象的值作为字符串插入，可以使用`{{.}}`。默认情况下，这个包使用`fmt`包来计算作为插入值的字符串。

要插入当前光标对象的某个字段的值，可以使用前缀为`.`的字段名称。例如，如果当前光标对象的类型为

```go
type Person struct {
        Name      string
        Age       int
        Emails     []string
        Jobs       []*Job
}

```

如下插入`Name`和`Age`的值:

```go
The name is {{.Name}}.
The age is {{.Age}}.

```

您可以使用`range`命令遍历数组或其他列表的元素。因此，要访问`Emails`数组的内容，您可以使用:

```go
{{range .Emails}}
        The email is {{.}}

{{end}}

```

在邮件循环期间，光标`.`被依次设置到每封邮件。循环结束时，光标返回到人。如果`Job`定义如下:

```go
type Job struct {
    Employer string
    Role     string
}

```

并且我们想要访问一个`person`的`jobs`的字段，我们可以像上面一样用一个`{{range .Jobs}}`来完成。另一种方法是将当前对象切换到`Jobs`字段。这是通过使用`{{with ...}} ... {{end}}`构造完成的，其中`{{.}}`是`Jobs`字段，这是一个数组:

```go
{{with .Jobs}}
    {{range .}}
        An employer is {{.Employer}}
        and the role is {{.Role}}
    {{end}}
{{end}}

```

您可以将它用于任何字段，而不仅仅是数组。

### 使用模板

一旦你有了一个模板，你就可以把它应用到一个对象来生成一个新的字符串，用这个对象来填充模板值。这是一个两步过程，包括解析模板，然后将其应用于对象。结果被输出到一个`Writer`，如:

```go
t := template.New("Person template")
t, err := t.Parse(templ)
if err == nil {
        buff := bytes.NewBufferString("")
        t.Execute(buff, person)
}

```

将模板应用到对象并打印到标准输出的示例程序是`PrintPerson.go`:

```go
/**
 * PrintPerson
 */

package main

import (
        "fmt"
        "text/template"
        "os"
)

type Person struct {
        Name   string
        Age    int
        Emails []string

        Jobs   []*Job
}

type Job struct {
        Employer string

        Role     string
}

const templ = `The name is {{.Name}}.
The age is {{.Age}}.
{{range .Emails}}
        An email is {{.}}
{{end}}

{{with .Jobs}}
    {{range .}}
        An employer is {{.Employer}}
        and the role is {{.Role}}
    {{end}}
{{end}}
`

func main() {
        job1 := Job{Employer: "Box Hill Institute", Role: "Director, Commerce and ICT"}
        job2 := Job{Employer: "Canberra University", Role: "Adjunct Professor"}

        person := Person{
                Name:   "jan",
                Age:    66,
                Emails: []string{"jan@newmarch.name", "jan.newmarch@gmail.com"},
                Jobs:   []*Job{&job1, &job2},
        }

        t := template.New("Person template")
        t, err := t.Parse(templ)
        checkError(err)

        err = t.Execute(os.Stdout, person)
        checkError(err)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

其输出如下所示:

```go
The name is jan.
The age is 66.

        An email is jan@newmarch.name

        An email is jan.newmarch@gmail.com

        An employer is Canberra University
        and the role is Adjunct Professor

        An employer is Box Hill Institute
        and the role is Director, Commerce and ICT

```

注意，在这个打印输出中有大量的空白作为换行符。这是因为我们在模板中有空白。如果您想减少这个空白，请删除模板中的换行符，如下所示:

```go
{{range .Emails}} An email is {{.}} {{end}}

```

另一种方法是使用命令分隔符`"{{- " and " -}}"`分别从紧接的前一个文本中删除所有尾随空白，从紧接的后一个文本中删除所有前导空白。

在这个例子中，我们在程序中使用了一个字符串作为模板。您也可以使用`template.ParseFiles()`功能从文件中加载模板。由于某种我不理解的原因(在早期版本中不需要)，分配给模板的名称必须与文件列表中第一个文件的基本名称相同。这是个 bug 吗？

## 管道

上述转换将文本片段插入到模板中。这些文本基本上是任意的，不管字段的字符串值是什么。如果我们希望它们作为 HTML 文档(或其他特殊形式)的一部分出现，我们必须对特定的字符序列进行转义。例如，要在 HTML 文档中显示任意文本，我们必须将`<`改为`&lt;`。Go 模板有许多内置函数，其中之一就是`html()`。这些函数以类似于 UNIX 管道的方式工作，从标准输入读取数据并写入标准输出。

要获取当前对象`.`的值并对其应用 HTML 转义，您需要在模板中编写一个“管道”:

```go
{{. | html}}

```

对于其他功能也是如此。

## 定义函数

模板使用对象的字符串表示来插入值，使用`fmt`包将对象转换成字符串。有时候这并不是我们所需要的。例如，为了避免垃圾邮件发送者获得电子邮件地址，很常见的是将符号`@`替换为单词“at”，如“jan at newmarch.name”。如果我们想使用模板以那种形式显示电子邮件地址，我们必须构建一个自定义函数来完成这种转换。

每个模板函数都有一个用于模板本身的名称和一个关联的 Go 函数。这些由以下类型链接:

```go
type FuncMap map[string]interface{}

```

例如，如果我们希望我们的模板函数是`emailExpand`，它链接到 Go 函数`EmailExpander`，我们将它添加到模板中的函数，如下所示:

```go
t = t.Funcs(template.FuncMap{"emailExpand": EmailExpander})

```

`EmailExpander`的签名通常如下:

```go
func EmailExpander(args ...interface{}) string

```

对于我们感兴趣的用法，函数应该只有一个参数，它将是一个字符串。Go 模板库中的现有函数有一些初始代码来处理不一致的情况，所以我们只是复制这些代码。然后，只需简单的字符串操作就可以改变电子邮件地址的格式。一个程序是`PrintEmails.go`:

```go
/**
 * PrintEmails
 */

package main

import (
        "fmt"
        "os"
        "strings"
        "text/template"
)

type Person struct {
        Name   string
        Emails []string
}

const templ = `The name is {{.Name}}.
{{range .Emails}}
        An email is "{{. | emailExpand}}"
{{end}}
`

func EmailExpander(args ...interface{}) string {

        ok := false
        var s string
        if len(args) == 1 {
                s, ok = args[0].(string)
        }
        if !ok {
                s = fmt.Sprint(args...)
        }

        // find the @ symbol
        substrs := strings.Split(s, "@")
        if len(substrs) != 2 {
                return s
        }
        // replace the @ by " at "
        return (substrs[0] + " at " + substrs[1])
}

func main() {
        person := Person{
                Name:   "jan",
                Emails: []string{"jan@newmarch.name", "jan.newmarch@gmail.com"},
        }

        t := template.New("Person template")

        // add our function
        t = t.Funcs(template.FuncMap{"emailExpand": EmailExpander})

        t, err := t.Parse(templ)

        checkError(err)

        err = t.Execute(os.Stdout, person)
        checkError(err)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

输出如下所示:

```go
The name is jan.

        An email is "jan at newmarch.name"

        An email is "jan.newmarch at gmail.com"

```

## 变量

模板包允许您定义和使用变量。作为这样做的动机，考虑我们如何打印每个人的电子邮件地址，并以他们的名字作为前缀。我们使用的类型也是这个:

```go
type Person struct {
        Name      string
        Emails     []string
}

```

为了访问电子邮件字符串，我们使用如下的`range`语句:

```go
{{range .Emails}}
    {{.}}
{{end}}

```

但是此时我们不能访问`Name`字段，因为`.`正在遍历数组元素，而`Name`不在这个范围内。解决方案是将`Name`字段的值保存在一个变量中，该变量在其作用域内的任何地方都可以被访问。模板中的变量以`$`为前缀。所以我们这样写:

```go
{{$name := .Name}}
{{range .Emails}}
    Name is {{$name}}, email is {{.}}
{{end}}

```

程序是`PrintNameEmails.go`:

```go
/**
 * PrintNameEmails
 */

package main

import (
        "text/template"
        "os"
        "fmt"
)

type Person struct {
        Name   string
        Emails []string
}

const templ = `{{$name := .Name}}
{{range .Emails}}
    Name is {{$name}}, email is {{.}}

{{end}}
`

func main() {
        person := Person{
                Name:   "jan",
                Emails: []string{"jan@newmarch.name", "jan.newmarch@gmail.com"},
        }

        t := template.New("Person template")
        t, err := t.Parse(templ)
        checkError(err)

        err = t.Execute(os.Stdout, person)
        checkError(err)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

以下是输出:

```go
Name is jan, email is jan@newmarch.name

Name is jan, email is jan.newmarch@gmail.com

```

## 条件语句

继续`Person`的例子，假设您只想打印出电子邮件列表，而不想深入研究它。您可以使用模板来做到这一点:

```go
Name is {{.Name}}
Emails are {{.Emails}}

```

这将打印以下内容:

```go
Name is jan
Emails are [jan@newmarch.name jan.newmarch@gmail.com]

```

因为这是`fmt`包显示列表的方式。

在许多情况下，这可能是好的，如果这是你想要的。让我们考虑一个几乎正确，但不完全正确的情况。有一个 JSON 包来序列化对象，我们在第 [4](04.html) 章中讨论过。这将产生以下结果:

```go
{"Name": "jan",
 "Emails": ["jan@newmarch.name", "jan.newmarch@gmail.com"]
}

```

JSON 包是您在实践中使用的包，但是让我们看看是否可以使用模板生成 JSON 输出。我们可以通过现有的模板做类似的事情。作为 JSON 序列化程序，这几乎是正确的:

```go
{"Name": "{{.Name}}",
 "Emails": {{.Emails}}

}

```

它会产生这样的结果:

```go
{"Name": "jan",
 "Emails": [jan@newmarch.name jan.newmarch@gmail.com]
}

```

这有两个问题:地址没有用引号括起来，列表元素应该用`,`分隔。

这样如何——看看数组元素，把它们放在引号中，然后加上逗号？

```go
{"Name": {{.Name}},
  "Emails": [
   {{range .Emails}}
      "{{.}}",
   {{end}}
  ]
}

```

这将产生:

```go
{"Name": "jan",
 "Emails": ["jan@newmarch.name", "jan.newmarch@gmail.com",]
}

```

(加上一些空格。)

同样，这几乎是正确的，但是如果仔细观察，您会看到在最后一个列表元素后面有一个尾随的`,`。根据 JSON 语法(参见 [`http://www.json.org/`](http://www.json.org/) )，这种尾随的`,`是不允许的。实现可能在处理这一问题的方式上有所不同。

我们想要的是打印除了最后一个元素之外的每个元素，后跟一个`,`。这实际上有点难做到，所以更好的方法是打印每个以 a 开头的元素，除了第一个。(这个技巧是我在栈溢出— [`http://stackoverflow.com/questions/201782/can-you-use-a-trailing-comma-in-a-json-object`](http://stackoverflow.com/questions/201782/can-you-use-a-trailing-comma-in-a-json-object) 从“brianb”那里得到的)。这更容易，因为第一个元素的索引是零，许多编程语言，包括 Go 模板语言，都将零视为布尔值`false`。

条件语句的一种形式是`{{if pipeline}} T1 {{else}} T0 {{end}}`。我们需要将`pipeline`作为电子邮件数组的索引。幸运的是，`range`声明的一个变体给了我们这个。有两种引入变量的形式:

```go
{{range $elmt := array}}
{{range $index, $elmt := array}}

```

所以我们通过数组建立了一个循环，如果索引是`false` (0)，我们就打印这个元素。否则，我们会在它前面打印一个`,`。模板如下所示:

```go
{"Name": "{{.Name}}",
 "Emails": [
 {{range $index, $elmt := .Emails}}
    {{if $index}}
        , "{{$elmt}}"
    {{else}}
         "{{$elmt}}"
    {{end}}
 {{end}}
 ]
}

```

完整的程序是`PrintJSONEmails.go`:

```go
/**
 * PrintJSONEmails
 */

package main

import (
        "text/template"
        "os"
        "fmt"
)

type Person struct {
        Name   string
        Emails []string
}

const templ = `{"Name": "{{.Name}}",
 "Emails": [
{{range $index, $elmt := .Emails}}
    {{if $index}}
        , "{{$elmt}}"
    {{else}}
         "{{$elmt}}"
    {{end}}
{{end}}
 ]
}
`

func main() {
        person := Person{
                Name:   "jan",
                Emails: []string{"jan@newmarch.name", "jan.newmarch@gmail.com"},
        }

        t := template.New("Person template")

        t, err := t.Parse(templ)
        checkError(err)

        err = t.Execute(os.Stdout, person)
        checkError(err)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

这给出了正确的 JSON 输出。

在离开本节之前，请注意使用逗号分隔符格式化列表的问题可以通过在 Go 中定义合适的函数来解决，这些函数可以作为模板函数使用。借用另一种编程语言中的一句名言，“有不止一种方法可以做到这一点！”。以下程序是罗杰·佩佩以`Sequence.go`的身份发给我的:

```go
/**
 * Sequence.go
 * Copyright Roger Peppe
 */

package main

import (
        "errors"
        "fmt"
        "os"
        "text/template"
)

var tmpl = `{{$comma := sequence "" ", "}}
{{range $}}{{$comma.Next}}{{.}}{{end}}
{{$comma := sequence "" ", "}}
{{$colour := cycle "black" "white" "red"}}
{{range $}}{{$comma.Next}}{{.}} in {{$colour.Next}}{{end}}
`

var fmap = template.FuncMap{
        "sequence": sequenceFunc,
        "cycle":    cycleFunc,
}

func main() {
        t, err := template.New("").Funcs(fmap).Parse(tmpl)
        if err != nil {
                fmt.Printf("parse error: %v\n", err)

                return
        }
        err = t.Execute(os.Stdout, []string{"a", "b", "c", "d", "e", "f"})
        if err != nil {
                fmt.Printf("exec error: %v\n", err)
        }
}

type generator struct {
        ss []string
        i  int
        f  func(s []string, i int) string
}

func (seq *generator) Next() string {
        s := seq.f(seq.ss, seq.i)
        seq.i++
        return s
}

func sequenceGen(ss []string, i int) string {
        if i >= len(ss) {
                return ss[len(ss)-1]
        }
        return ss[i]
}

func cycleGen(ss []string, i int) string {
        return ss[i%len(ss)]
}

func sequenceFunc(ss ...string) (*generator, error) {
        if len(ss) == 0 {
                return nil, errors.New("sequence must have at least one element")
        }
        return &generator{ss, 0, sequenceGen}, nil
}

func cycleFunc(ss ...string) (*generator, error) {
        if len(ss) == 0 {
                return nil, errors.New("cycle must have at least one element")
        }
        return &generator{ss, 0, cycleGen}, nil
}

```

以下是输出:

```go
a, b, c, d, e, f

a in black, b in white, c in red, d in black, e in white, f in red

```

## HTML/模板包

前面的程序都处理了`text/template`包。这将应用转换，而不考虑可能使用文本的任何上下文。例如，如果`PrintPerson.go`中的文本变为:

```go
job1 := Job{Employer: "<script>alert('Could be nasty!')</script>", Role: "Director, Commerce and ICT"}

```

该程序将生成以下文本:

```go
An employer is <script>alert('Could be nasty!')</script>

```

如果下载到浏览器中，将会产生意想不到的效果。

在管道中使用`html`命令可以减少这种情况，如{{。| html}}，并将产生以下内容:

```go
An employer is &lt;script&gt;alert('Could be nasty!')&lt;/script&gt

```

将此过滤器应用于所有表达式将变得繁琐。此外，它可能无法捕捉潜在危险的 JavaScript、CSS 或 URI 表达式。

`html/template`包就是为了克服这些问题而设计的。通过用`html/template`替换`text/template`的简单步骤，适当的转换将被应用到结果文本，净化它，使它适合 web 上下文。

## 结论

Go 模板包对于某些涉及插入对象值的文本转换非常有用。例如，它不具备正则表达式的能力，但它比正则表达式更快，在许多情况下也更容易使用。