# 十一、HTML

Web 最初是为 HTML 文档服务而创建的。现在它被用来服务各种各样的文件和不同种类的数据。然而，HTML 仍然是通过网络传递的主要文档类型。

HTML 经历了大量的版本，目前的版本是 HTML5。也有许多 HTML 的“供应商”版本，引入了从未成为标准的标签。

HTML 非常简单，可以手工编辑。因此，许多 HTML 文档是“格式错误的”，这意味着它们不遵循语言的语法。HTML 解析器通常不是很严格，会接受许多“非法”文档。

HTML 包本身只有两个功能——`EscapeString`和`UnescapeString`。这些适当地处理角色，例如`<`，将它们转换成`&lt;`，然后再转换回来。

这种方法的主要用途可能是对 HTML 文档中的标记进行转义，这样如果在浏览器中显示，就会显示所有的标记(很像 Linux 上 Chrome 中的 Ctrl+U 或 Mac Chrome 上的 Option+Cmd+U)。

我更倾向于用这个把程序的文本显示成网页。大多数编程语言都有`<`符号，许多有`&`。除非正确转义，否则这些会搞乱 HTML 查看器。我喜欢直接从文件系统中显示程序文本，而不是复制粘贴到文档中，以避免不同步。

下面的程序`EscapeString.go`是一个 web 服务器，它以预格式化的代码显示其 URL，并对麻烦的字符进行了转义:

```go
/*
 * This program serves a file in preformatted, code layout
 * Useful for showing program text, properly escaping special
 * characters like '<', '>' and '&'
 */

package main

import (
        "fmt"
        "html"
        "io/ioutil"
        "net/http"
        "os"
)

func main() {
        http.HandleFunc("/", escapeString)

        err := http.ListenAndServe(":8080", nil)
        checkError(err)

}

func escapeString(rw http.ResponseWriter, req *http.Request) {
        fmt.Println(req.URL.Path)
        bytes, err := ioutil.ReadFile("." + req.URL.Path)
        if err != nil {
                rw.WriteHeader(http.StatusNotFound)
                return
        }

        escapedStr := html.EscapeString(string(bytes))
        htmlText := "<html><body><pre><code>" +
                escapedStr +
                " </code></pre></body></html>"
        rw.Write([]byte(htmlText))
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Error ", err.Error())
                os.Exit(1)
        }
}

```

当它运行时，从包括`EscapeString.go`程序的目录中提供文件，浏览器将使用 URL `localhost:8080/EscapeString.go`正确显示它。

使用以下命令运行服务器:

```go
go run EscapeString.go

```

例如，使用以下命令运行客户端:

```go
curl localhost:8080/EscapeString.go

```

## Go HTML/模板包

对 web 服务器的攻击有很多种，其中最著名的是 SQL 注入，用户代理将数据输入到 web 表单中，该表单被故意设计为传递到数据库并在那里造成严重破坏。Go 没有任何特定的支持来避免这种情况，因为对于可以成功的 SQL 注入技术，数据库之间存在许多差异。SQL 注入预防备忘单(参见 [`https://www.owasp.org/index.php/SQL_Injection_Prevention_Cheat_Sheet`](https://www.owasp.org/index.php/SQL_Injection_Prevention_Cheat_Sheet) )总结了针对此类攻击的防御措施。主要的一点是通过使用 SQL 预准备语句来避免这种攻击，这可以通过使用`database/sql`包中的`Prepare`函数来完成。

更微妙的攻击是基于 XSS——跨站点脚本。在这种情况下，攻击者不是试图攻击网站本身，而是在服务器上存储恶意代码，以攻击该网站的任何客户端。

这些攻击基于将数据插入到数据库字符串中，例如，当将数据传送到浏览器时，将攻击浏览器，并通过它攻击网站的客户端。(这有几种变体，在“OWASP:跨站点脚本的类型”——[`https://www.owasp.org/index.php/Types_of_Cross-Site_Scripting`](https://www.owasp.org/index.php/Types_of_Cross-Site_Scripting)中讨论过)。)

例如，可以在请求博客评论的地方插入 JavaScript，以将浏览器重定向到攻击者的站点:

```go
<script>
   window.location='http://attacker/'
</script>

```

Go `html/` `template`包设计在`text/template`包之上。假设模板是可信的，但它处理的数据可能不可信。`html/template`增加的是数据的适当转义，以尽量消除 XSS 的可能性。它基于由 Mike Samuel 和 Prateek Saxena 撰写的名为“使用类型推断使 Web 模板抵抗 XSS”的文档。请在 [`https://rawgit.com/mikesamuel/sanitized-jquery-templates/trunk/safetemplate.html#problem_definition`](https://rawgit.com/mikesamuel/sanitized-jquery-templates/trunk/safetemplate.html#problem_definition) 阅读该论文，了解软件包背后理论以及软件包文档本身。

简而言之，按照`text/template`包准备模板，如果结果文本被交付给 HTML 代理，则使用`html/template`包。

## 标记 HTML

Go 子仓库中的包`golang.org/x/net/html`包含一个 HTML 标记器。这允许您构建 HTML 标记的解析树。它符合 HTML5。

运行以下命令后可以使用它:

```go
go get golang.org/x/net/html

```

使用它的一个示例程序是`ReadHTML.go`:

```go
/* Read HTML
 */

package main

import (
        "fmt"
        "golang.org/x/net/html"

        "io/ioutil"
        "os"
        "strings"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "file")
                os.Exit(1)
        }
        file := os.Args[1]
        bytes, err := ioutil.ReadFile(file)
        checkError(err)
        r := strings.NewReader(string(bytes))

        z := html.NewTokenizer(r)

        depth := 0
        for {
                tt := z.Next()

                for n := 0; n < depth; n++ {
                        fmt.Print(" ")
                }

                switch tt {
                case html.ErrorToken:
                        fmt.Println("Error ", z.Err().Error())
                        os.Exit(0)
                case html.TextToken:
                        fmt.Println("Text: \"" + z.Token().String() + "\"")
                case html.StartTagToken, html.EndTagToken:
                        fmt.Println("Tag: \"" + z.Token().String() + "\"")
                        if tt == html.StartTagToken {
                                depth++
                        } else {
                                depth--
                        }

                }
        }

}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

当它在一个简单的 HTML 文档上运行时，如下所示:

```go
<html>
  <head>
    <title> Test HTML </title>
  </head>
  <body>
    <h1> Header one </h1>
    <p>
      Test para
    </p>
  </body>
</html>

```

它产生以下内容:

```go
Tag: "<html>"
 Text: "
  "
 Tag: "<head>"
  Text: "
    "
  Tag: "<title>"
   Text: " Test HTML "
   Tag: "</title>"
  Text: "
  "
  Tag: "</head>"
 Text: "
  "
 Tag: "<body>"
  Text: "
    "
  Tag: "<h1>"
   Text: " Header one "
   Tag: "</h1>"
  Text: "
    "
  Tag: "<p>"
   Text: "
      Test para
    "
   Tag: "</p>"
  Text: "
  "
  Tag: "</body>"

 Text: "
"
 Tag: "</html>"
Text: "
"

```

(它产生的所有空白都是正确的。)

## XHTML/HTML

XML 包中对 XHTML/HTML 的支持也是有限的，这将在下一章讨论。

## 数据

JSON 有很好的支持，如第 4 章所讨论的。

## 结论

这个包裹没什么特别的。关于模板的第 [9](09.html) 章讨论了子包`html/template`。