# 十二、XML

XML 是一种重要的标记语言，主要用于以文本格式表示结构化数据。在我们在第 [4](04.html) 章使用的语言中，它可以被认为是将数据结构序列化为文本文档的一种手段。它用于描述 DocBook 和 XHTML 等文档。它用于专门的标记语言，如 MathML 和 CML(化学标记语言)。它用于将数据编码为 Web 服务的 SOAP 消息，并且可以使用 WSDL (Web 服务描述语言)来指定 Web 服务。

在最简单的层面上，XML 允许您定义自己的标签，以便在文本文档中使用。标签可以嵌套，也可以穿插文本。每个标签还可以包含带值的属性。例如，文件`person.xml`可能包含:

```go
<person>
  <name>
    <family> Newmarch </family>
    <personal> Jan </personal>
  </name>
  <email type="personal">
    jan@newmarch.name
  </email>
  <email type="work">
    j.newmarch@boxhill.edu.au
  </email>
</person>

```

任何 XML 文档的结构都可以用多种方式描述:

*   文档类型定义 DTD 有利于描述结构
*   XML 模式适合描述 XML 文档使用的数据类型
*   RELAX NG 被提议作为两者的替代方案

对于定义 XML 文档结构的每种方法的相对价值存在争议。我们不会买那个，因为 Go 不支持任何一个。Go 不能根据模式检查任何文档的有效性，只能检查文档的格式是否良好。甚至良构性也是 XML 文档的一个重要特征，并且在实践中经常是 HTML 文档的一个问题。这使得 XML 适合于表示非常复杂的数据，而 HTML 不适合。

本章讨论了四个主题:解析 XML 流、将 Go 数据编组和解组成 XML 以及 XHTML。

## 解析 XML

Go 有一个 XML 解析器，它是使用来自`encoding/xml`包的`NewDecoder`创建的。这将一个`io.Reader`作为参数，并返回一个指向`Decoder`的指针。这种类型的主要方法是`Token`，它返回输入流中的下一个令牌。令牌是这些类型之一— `StartElement`、`EndElement`、`CharData`、`Comment`、`ProcInst`或`Directive`。

我们将使用这种类型:

```go
type Name struct {
    Space, Local string
}

```

XML 类型有`StartElement`、`EndElement`、`CharData`、`Comment`、`ProcInst`和`Directive`。接下来将对它们进行描述。

### startellemon 类型

类型`StartElement`是具有两种字段类型的结构:

```go
type StartElement struct {
    Name Name
    Attr []Attr
}

```

在哪里

```go
type Attr struct {
    Name  Name
    Value string
}

```

### EndElement 类型

这也是如下的结构:

```go
type EndElement struct {
    Name Name
}

```

### CharData 类型

这种类型表示由标记括起的文本内容，是一种简单类型:

```go
type CharData []byte

```

### 注释类型

类似地，对于这种类型:

```go
type Comment []byte

```

### ProcInst 类型

A `ProcInst`表示形式为`<?target inst?>`的 XML 处理指令:

```go
type ProcInst struct {
    Target string
    Inst   []byte
}

```

### 指令类型

一个`Directive`表示一个形式为`<!text>`的 XML 指令。这些字节不包括`<!`和`>`标记。

```go
type Directive []byte

```

打印出 XML 文档树形结构的程序是`ParseXML.go`:

```go
/* Parse XML
 */

package main

import (
        "encoding/xml"
        "fmt"
        "io/ioutil"
        "os"
        "strings"
)

func main() {
        if len(os.Args) != 2 {
                fmt.Println("Usage: ", os.Args[0], "file")
                os.Exit(1)
        }
        file := os.Args[1]
        bytes, err := ioutil.ReadFile(file)
        checkError(err)
        r := strings.NewReader(string(bytes))

        parser := xml.NewDecoder(r)
        depth := 0
        for {
                token, err := parser.Token()
                if err != nil {
                        break

                }
                switch t := token.(type) {
                case xml.StartElement:
                        elmt := xml.StartElement(t)
                        name := elmt.Name.Local
                        printElmt(name, depth)
                        depth++
                case xml.EndElement:
                        depth--
                        elmt := xml.EndElement(t)
                        name := elmt.Name.Local
                        printElmt(name, depth)
                case xml.CharData:
                        bytes := xml.CharData(t)
                        printElmt("\""+string([]byte(bytes))+"\"", depth)
                case xml.Comment:
                        printElmt("Comment", depth)
                case xml.ProcInst:
                        printElmt("ProcInst", depth)
                case xml.Directive:
                        printElmt("Directive", depth)
                default:
                        fmt.Println("Unknown")
                }
        }
}

func printElmt(s string, depth int) {
        for n := 0; n < depth; n++ {
                fmt.Print("  ")
        }
        fmt.Println(s)
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

注意，解析器包含所有的`CharData`，包括标签之间的空白。

如果我们针对前面给出的`person`数据结构运行这个程序，如下所示:

```go
go run ParseXML.go person.xml

```

它产生以下内容:

```go
person
  "
  "
  name
    "
    "
    family
      " Newmarch "
    family
    "
    "
    personal
      " Jan "
    personal
    "
  "
  name
  "
  "
  email
    "
    jan@newmarch.name
  "
  email
  "
  "
  email
    "
    j.newmarch@boxhill.edu.au
  "
  email
  "
"
person
"
"

```

注意，因为没有使用 DTD 或其他 XML 规范，所以标记器正确地打印出了所有的空白(DTD 可能指定空白可以忽略，但是没有它就不能做出这样的假设)。

使用这个解析器有一个潜在的陷阱。它为字符串重用空间，因此一旦看到一个令牌，如果以后想引用它，就需要复制它的值。Go 有像`func (c CharData) Copy() CharData`这样的方法来制作数据的副本。

## 解组 XML

Go 提供了一个名为`Unmarshal`的函数来将 XML 解组到 Go 数据结构中。解组并不完美:Go 和 XML 是不同的语言。

在看细节之前，我们先考虑一个简单的例子。首先考虑前面给出的 XML 文档:

```go
<person>
  <name>
    <family> Newmarch </family>
    <personal> Jan </personal>
  </name>
  <email type="personal">
    jan@newmarch.name
  </email>
  <email type="work">
    j.newmarch@boxhill.edu.au
  </email>
</person>

```

我们希望将其映射到 Go 结构中:

```go
type Person struct {
        Name Name
        Email []Email
}

type Name struct {
        Family string
        Personal string
}

type Email struct {
        Type string
        Address string
}

```

这需要几点说明:

*   解组使用 Go 反射包。这要求所有字段都是公共的，即以大写字母开头。早期版本的 Go 使用不区分大小写的匹配来匹配字段，比如 XML 字符串“name”和字段`Name`。不过，现在使用的是区分大小写的匹配。要执行匹配，必须标记结构字段，以显示将要匹配的 XML 字符串。这将`Person`更改为以下内容:

    ```go
    type Person struct {
            Name Name `xml:"name"`
            Email []Email `xml:"email"`
    }

    ```

*   虽然对字段进行标记可以将 XML 字符串附加到字段上，但它不能对结构名进行标记。需要一个附加字段，字段名为`XMLName`。这只影响顶级结构，`Person` :

    ```go
    type Person struct {
            XMLName Name `xml:"person"`
            Name Name `xml:"name"`
            Email []Email `xml:"email"`
    }

    ```

*   重复的标签映射到 Go 中的一个切片。
*   只有当 Go 字段具有标签`,attr`时，标签中的属性才会与结构中的字段匹配。这发生在`Email`的字段`Type`中，其中匹配`email`标签的属性`type`需要`xml:"type,attr"`。
*   如果一个 XML 标签没有属性，只有字符数据，那么它匹配一个同名的`string`字段(尽管区分大小写)。因此带有字符数据`Newmarch`的标签`xml:"family"`映射到字符串字段`Family`。
*   但是如果标签有属性，那么它必须映射到一个结构。Go 将字符数据分配给标签为`,chardata`的字段。这发生在`email`数据和`tag ,chardata`字段`Address`中。

解组上面文档的程序是`Unmarshal.go`:

```go
/* Unmarshal
 */

package main

import (
        "encoding/xml"
        "fmt"
        "os"
)

type Person struct {
        XMLName Name    `xml:"person"`
        Name    Name    `xml:"name"`
        Email   []Email `xml:"email"`
}

type Name struct {
        Family   string `xml:"family"`
        Personal string `xml:"personal"`
}

type Email struct {
        Type    string `xml:"type,attr"`
        Address string `xml:",chardata"`
}

func main() {
        str := `<?xml version="1.0" encoding="utf-8"?>
<person>
  <name>
    <family> Newmarch </family>
    <personal> Jan </personal>
  </name>
  <email type="personal">
    jan@newmarch.name
  </email>
  <email type="work">
    j.newmarch@boxhill.edu.au
  </email>
</person>`

        var person Person

        err := xml.Unmarshal([]byte(str), &person)
        checkError(err)

        // now use the person structure e.g.
        fmt.Println("Family name: \"" + person.Name.Family + "\"")
        fmt.Println("Second email address: \"" + person.Email[1].Address + "\"")
}

func checkError(err error) {
        if err != nil {
                fmt.Println("Fatal error ", err.Error())
                os.Exit(1)
        }
}

```

(注意空格是正确的。)封装规范中给出了严格的规则。

## 编组 XML

Go 还支持将数据结构组织成 XML 文档。功能是:

```go
func Marshal(v interface}{) ([]byte, error)

```

整理一个简单结构的程序是`Marshal.go`:

```go
/* Marshal
 */

package main

import (
        "encoding/xml"
        "fmt"
)

type Person struct {
        Name  Name
        Email []Email
}

type Name struct {
        Family   string
        Personal string
}

type Email struct {
        Kind    string "attr"
        Address string "chardata"
}

func main() {
        person := Person{
                Name: Name{Family: "Newmarch", Personal: "Jan"},
                Email: []Email{Email{Kind: "home", Address: "jan"},
                        Email{Kind: "work", Address: "jan"}}}

        buff, _ := xml.Marshal(person)
        fmt.Println(string(buff))
}

```

它生成不带空格的文本:

```go
<Person><Name><Family>Newmarch</Family><Personal>Jan</Personal></Name><Email><Kind>home</Kind><Address>jan</Address></Email><Email><Kind>work</Kind><Address>jan</Address></Email></Person>

```

## 可扩展的超文本标记语言

HTML 不符合 XML 语法。它有未终止的标签，如`<br>`。XHTML 是对 HTML 的清理，使其符合 XML。XHTML 中的文档可以使用上述 XML 技术进行管理。XHTML 似乎没有像最初预期的那样被广泛使用。我个人的怀疑是，HTML 解析器通常是容忍错误的，当在浏览器中使用时，通常可以合理地呈现文档，即使在浏览器中，XHTML 解析器也往往更加严格，经常在遇到甚至一个 XML 错误时也不能呈现任何内容。对于面向用户的软件来说，这通常不是合适的行为。

## 超文本标记语言

XML 包中有一些处理 HTML 文档的支持，即使它们可能不符合 XML。如果关闭严格的解析检查，前面讨论的 XML 解析器可以处理许多 HTML 文档。

```go
parser := xml.NewDecoder(r)
parser.Strict = false
parser.AutoClose = xml.HTMLAutoClose
parser.Entity = xml.HTMLEntity

```

## 结论

Go 具有处理 XML 字符串的基本支持。它还没有处理 XML 规范语言(如 XML Schema 或 Relax NG)的机制。